# Copyright (C) 2024 Peter Ropač
# This file is part of SPOMSO.
# SPOMSO is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# SPOMSO is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public License along with SPOMSO. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from spomso.cores.geom import VectorField
from spomso.cores.vector_functions import cartesian_define, cylindrical_define, spherical_define
from spomso.cores.vector_functions import radial_vector_field_cylindrical, radial_vector_field_spherical
from spomso.cores.vector_functions import hyperbolic_vector_field_cylindrical, awn_vector_field_cylindrical
from spomso.cores.vector_functions import aar_vector_field_cylindrical
from spomso.cores.vector_functions import vortex_vector_field_cylindrical, aav_vector_field_cylindrical
from spomso.cores.vector_functions import x_vector_field, y_vector_field, z_vector_field
from spomso.cores.vector_functions import from_sdf


class CartesianVectorField(VectorField):
    """
    Vector field defined by its components in the cartesian coordinate system (x, y, z).
    The components of the output vector field are cartesian (x, y, z).
    """
    def __init__(self):
        VectorField.__init__(self, cartesian_define)


class CylindricalVectorField(VectorField):
    """
    Vector field defined by its components in the cylindrical coordinate system (r, phi, z).
    The components of the output vector field are cartesian (x, y, z).
    """
    def __init__(self):
        VectorField.__init__(self, cylindrical_define)


class SphericalVectorField(VectorField):
    """
    Vector field defined by its components in the spherical coordinate system (r, phi, theta).
    The components of the output vector field are cartesian (x, y, z).
    """
    def __init__(self):
        VectorField.__init__(self, spherical_define)


class RadialSphericalVectorField(VectorField):
    """
    Vector field where all the vectors are pointing radially outwards from the origin.
    Point cloud specifying the positions of points at which the vector field is evaluated is taken as the input.
    The components of the output vector field are cartesian (x, y, z).
    """
    def __init__(self):
        VectorField.__init__(self, radial_vector_field_spherical)


class RadialCylindricalVectorField(VectorField):
    """
    Vector field where all the vectors are pointing radially outwards from the line x=0, y=0.
    Point cloud specifying the positions of points at which the vector field is evaluated is taken as the input.
    The components of the output vector field are cartesian (x, y, z).
    """
    def __init__(self):
        VectorField.__init__(self, radial_vector_field_cylindrical)


class HyperbolicCylindricalVectorField(VectorField):
    """
    Hyperbolic vector field centered at the line x=0, y=0.
    Point cloud specifying the positions of points at which the vector field is evaluated is taken as the input.
    The components of the output vector field are cartesian (x, y, z).
    """
    def __init__(self):
        VectorField.__init__(self, hyperbolic_vector_field_cylindrical)


class WindingCylindricalVectorField(VectorField):
    """
    Vector field where the vectors rotate around the line x=0, y=0, based on the winding number.
    Point cloud specifying the positions of points at which the vector field is evaluated is taken as the input.
    The components of the output vector field are cartesian (x, y, z).

    Args:
        gamma: winding number.
    """
    def __init__(self, gamma: float | int | np.ndarray):
        self._gamma = gamma
        VectorField.__init__(self, awn_vector_field_cylindrical, (gamma, ))

    @property
    def gamma(self) ->  float | int | np.ndarray:
        """ Winding number.

        Returns:
            Value of the set winding number.
        """
        return self._gamma


class AngledRadialCylindricalVectorField(VectorField):
    """
    Modified radial vector field centered at the line x=0, y=0.

    The vectors are at the specified angle with respect to the lines pointing outward from the line x=0, y=0.
    Point cloud specifying the positions of points at which the vector field is evaluated is taken as the input.

    The components of the output vector field are cartesian (x, y, z).

    Args:
        alpha: Angle between the vectors and the equidistant surfaces from the line x=0, y=0.
    """
    def __init__(self, alpha: float | int | np.ndarray):
        self._alpha = alpha
        VectorField.__init__(self, aar_vector_field_cylindrical, (alpha, ))

    @property
    def alpha(self) -> float | int | np.ndarray:
        """ Angle between the vectors and the equidistant surfaces from the line x=0, y=0.

        Returns:
            Value of alpha.
        """
        return self._alpha


class VortexCylindricalVectorField(VectorField):
    """
    Vortex vector field centered at the line x=0, y=0.
    Point cloud specifying the positions of points at which the vector field is evaluated is taken as the input.
    The components of the output vector field are cartesian (x, y, z).
    """
    def __init__(self):
        VectorField.__init__(self, vortex_vector_field_cylindrical)


class AngledVortexCylindricalVectorField(VectorField):
    """
    Modified vortex vector field centered at the line x=0, y=0.
    The vectors are at the specified angle with respect to the equidistant surfaces from the line x=0, y=0.
    Point cloud specifying the positions of points at which the vector field is evaluated is taken as the input.
    The components of the output vector field are cartesian (x, y, z).

    Args:
        alpha: Angle between the vectors and the equidistant surfaces from the line x=0, y=0.
    """
    def __init__(self, alpha: float | int | np.ndarray):
        self._alpha = alpha
        VectorField.__init__(self, aav_vector_field_cylindrical, (alpha, ))

    @property
    def alpha(self) -> float | int | np.ndarray:
        """ Angle between the vectors and the equidistant surfaces from the line x=0, y=0.

        Returns:
            Value of alpha.
        """
        return self._alpha


class XVectorField(VectorField):
    """
    Vector field where only the X component (cartesian coordinates) is non-zero.
    Point cloud specifying the positions of points at which the vector field is evaluated is taken as the input.
    The components of the output vector field are cartesian (x, y, z).
    """
    def __init__(self):
        VectorField.__init__(self, x_vector_field)


class YVectorField(VectorField):
    """
    Vector field where only the Y component (cartesian coordinates) is non-zero.
    Point cloud specifying the positions of points at which the vector field is evaluated is taken as the input.
    The components of the output vector field are cartesian (x, y, z).
    """
    def __init__(self):
        VectorField.__init__(self, y_vector_field)


class ZVectorField(VectorField):
    """
    Vector field where only the Z component (cartesian coordinates) is non-zero.
    Point cloud specifying the positions of points at which the vector field is evaluated is taken as the input.
    The components of the output vector field are cartesian (x, y, z).
    """
    def __init__(self):
        VectorField.__init__(self, z_vector_field)


class VectorFieldFromSDF(VectorField):
    """
    Vector field constructed from an SDF.
    Point cloud specifying the value of the SDF is taken as an input.
    The components of the output vector field are cartesian (x, y, z).

    Args:
        grid_resolution: Number of points along each axis in the grid on which the SDF is evaluated.
    """
    def __init__(self, grid_resolution: tuple | list | np.ndarray):
        VectorField.__init__(self, from_sdf,  grid_resolution)








# Copyright (C) 2024 Peter Ropač
# This file is part of SPOMSO.
# SPOMSO is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# SPOMSO is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public License along with SPOMSO. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from scipy.interpolate import NearestNDInterpolator
from scipy.spatial.distance import cdist
from scipy.spatial import KDTree


def sdf_circle(co, radius):
    length = np.linalg.norm(co[:2], axis=0)
    return length - radius


def sdf_box_2d(co, size):
    d = np.subtract(np.abs(co[:2, :]).T, size).T

    term1 = np.linalg.norm(np.maximum(d, 0), axis=0)
    term2 = np.minimum(np.maximum(d[0],d[1]), 0)
    return term1 + term2


def sdf_segment_2d(co, a, b):
    a = np.asarray(a[:2])
    b = np.asarray(b[:2])
    pa = np.subtract(co[:2, :].T, a).T
    ba = b - a
    h = np.clip(np.sum(np.multiply(pa.T, ba).T, axis=0)/np.dot(ba, ba), 0, 1)
    length = np.linalg.norm(pa - np.outer(ba,h), axis=0)
    return length


def sdf_rounded_box_2d(co, size, rounding):

    m1 = co[0] > 0
    m2 = co[1] > 0
    m3 = co[0] < 0

    r = rounding[0]*np.ones(co.shape[1])
    r[m1] = rounding[1]
    r[m2] = rounding[2]
    r[m3*m2] = rounding[3]

    d = np.subtract(np.abs(co[:2, :]).T, size).T + r
    o = np.linalg.norm(np.maximum(d, 0), axis=0)
    u = np.minimum(np.maximum(d[0], d[1]), 0) - r

    return o + u


def sdf_triangle_2d(co, p0, p1, p2):

    e0 = p1 - p0
    e1 = p2 - p1
    e2 = p0 - p2

    v0 = np.subtract(co[:2,:].T, p0).T
    v1 = np.subtract(co[:2,:].T, p1).T
    v2 = np.subtract(co[:2,:].T, p2).T

    pq0 = v0 - np.outer(e0, np.clip(np.dot(v0.T, e0).T / np.dot(e0, e0), 0, 1))
    pq1 = v1 - np.outer(e1, np.clip(np.dot(v1.T, e1).T / np.dot(e1, e1), 0, 1))
    pq2 = v2 - np.outer(e2, np.clip(np.dot(v2.T, e2).T / np.dot(e2, e2), 0, 1))

    s = np.sign(e0[0]*e2[1] - e0[1]*e2[0])

    dv0 = np.asarray( ( np.sum(np.multiply(pq0, pq0), axis=0), s*(v0[0]*e0[1]-v0[1]*e0[0]) ) )
    dv1 = np.asarray((np.sum(np.multiply(pq1, pq1), axis=0), s * (v1[0] * e1[1] - v1[1] * e1[0])))
    dv2 = np.asarray((np.sum(np.multiply(pq2, pq2), axis=0), s * (v2[0] * e2[1] - v2[1] * e2[0])))

    d = np.amin([dv0, dv1, dv2], axis=0)

    return -np.sqrt(d[0])*np.sign(d[1])


def sdf_arc_positive_only(co, radius, start_angle, end_angle):

    co = co[:2]

    phi = np.arctan2(co[1], co[0])
    phi[phi < 0] = 2 * np.pi + phi[phi < 0]

    ad = end_angle - start_angle
    h = np.clip((phi - start_angle)/ad, 0, 1)
    psi = ad*h + start_angle
    oc = np.asarray((radius*np.cos(psi),
                    radius*np.sin(psi) ))

    poc = np.subtract(co, oc)

    length = np.linalg.norm(poc, axis=0)
    start_length = np.linalg.norm(np.subtract(co.T, (radius*np.cos(start_angle), radius*np.sin(start_angle))).T, axis=0)
    return np.minimum(length, start_length)


def sdf_arc(co, radius, start_angle, end_angle):

    co = co[:2]
    c_angle = (start_angle + end_angle)/2
    rot = np.asarray([[np.cos(c_angle), np.sin(c_angle)], [-np.sin(c_angle), np.cos(c_angle)]])
    co = rot.dot(co)
    co[1] = np.abs(co[1])
    phi = np.arctan2(co[1], co[0])

    end_angle = np.abs(end_angle - c_angle)

    psi = np.clip(phi, 0, end_angle)
    oc = np.asarray((radius*np.cos(psi),
                    radius*np.sin(psi) ))

    poc = np.subtract(co, oc)

    length = np.linalg.norm(poc, axis=0)
    return length


def sdf_sector_old(co, radius, angle):
    # UNFINISHED
    c = np.asarray([np.sin(angle), np.cos(angle)])
    co = co[:2]
    co[0] = np.abs(co[0])

    l = np.linalg.norm(co, axis=0) - radius
    m = np.linalg.norm(co - np.outer(c, np.clip( np.dot(co.T, c), 0, radius)))

    return np.maximum(l, m*np.sign(c[1]*co[0] - c[0]*co[1]))


def sdf_sector(co, radius, angle_1, angle_2):

    angle_diff = np.abs(angle_2 - angle_1)/2
    c_angle = (angle_2 + angle_1)/2

    rot = np.asarray([[np.cos(c_angle), np.sin(c_angle)], [-np.sin(c_angle), np.cos(c_angle)]])
    co = rot.dot(co[:2])
    co[1] = np.abs(co[1])

    phi = np.arctan2(co[1], co[0])
    psi = np.clip(phi, 0, angle_diff)
    oc = np.asarray((radius * np.cos(psi),
                     radius * np.sin(psi)))
    poc = np.subtract(co, oc)
    length = np.linalg.norm(poc, axis=0)

    c = np.asarray([np.cos(angle_diff), np.sin(angle_diff)])
    outer = np.outer(c, np.clip( np.dot(co.T, c).T, 0, radius))
    m = np.linalg.norm(co - outer, axis=0)

    msk = (np.linalg.norm(co, axis=0) <= radius)*(phi <= angle_diff)
    out = np.minimum(m,length)
    out[msk] = -out[msk]

    return out


def sdf_inf_sector(co, angle_1, angle_2):

    angle_diff = np.abs(angle_2 - angle_1)/2
    c_angle = (angle_2 + angle_1)/2

    rot = np.asarray([[np.cos(c_angle), np.sin(c_angle)], [-np.sin(c_angle), np.cos(c_angle)]])
    co = rot.dot(co[:2])
    co[1] = np.abs(co[1])

    phi = np.arctan2(co[1], co[0])

    c = np.asarray([np.cos(angle_diff), np.sin(angle_diff)])
    outer = np.outer(c, np.clip( np.dot(co.T, c).T, 0, np.inf))
    m = np.linalg.norm(co - outer, axis=0)

    sign = (phi - angle_diff)
    out = sign*m

    return out


def sdf_ngon(co, radius, n):

    co = co[:2, :]
    beta = np.pi*(0.5 - 1/n)
    alpha = 2*np.pi/n

    phi = np.arctan2(co[1], co[0])
    phi[phi<0] = 2*np.pi + phi[phi<0]
    phi = np.mod(phi, alpha)
    r_ = np.linalg.norm(co, axis=0)
    co = np.asarray([np.cos(phi), np.sin(phi)])*r_

    qo = np.subtract(co.T, [radius, 0]).T
    s = np.sin(beta)
    c = np.cos(beta)
    t = np.asarray([-c, s])
    no = np.asarray([s, c])
    l = 2*radius*np.sin(alpha/2)

    h = np.clip(np.dot(qo.T, t), 0, l)
    length = np.linalg.norm(qo - np.outer(t, h), axis=0)
    sign = np.sign(np.dot(qo.T, no))
    out = length*sign

    return out


def sdf_segmented_curve_2d(co, points, t):

    v = np.floor(t).astype(int)
    u = t - v
    fval = points[:2, v+1]*u + points[:2, v]*(1-u)

    tree = KDTree(fval.T)
    mindist, minid = tree.query(co[:2, :].T)
    return mindist


def sdf_segmented_line_2d(co, points):

    out = np.ones(co.shape[1])*10000
    for i in range(points.shape[1]-1):
        part = sdf_segment_2d(co, points[:, i], points[:, i+1])
        out = np.minimum(out, part)

    return out


def sdf_parametric_curve_2d(co, f, f_parameters, t):
    fval = f(t, *f_parameters)
    tree = KDTree(fval.T)
    mindist, minid = tree.query(co[:2, :].T)
    return mindist


def sdf_point_cloud_2d(co, points):
    tree = KDTree(points[:2, :].T)
    mindist, minid = tree.query(co[:2, :].T)
    return mindist




# Copyright (C) 2024 Peter Ropač
# This file is part of SPOMSO.
# SPOMSO is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# SPOMSO is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public License along with SPOMSO. If not, see <https://www.gnu.org/licenses/>.


from spomso.cores.geom import VectorField
from spomso.cores.vector_functions_special import lcwg1_2d, lcwg1_m1, lcwg1_p1


class LCWG2D(VectorField):

    def __init__(self,  parameters, co_resolution, sign):
        """
        Vector field of a Liquid Crystal waveguide.
        The vector field is generated based on the distance field and size parameters of the waveguide.
        Point cloud specifying the values of the Signed Distance Function is taken as the input.
        The vector field is independent of the z coordinate (cartesian coordinates).
        The components of the output vector field are cartesian (x, y, z).
        :param parameters: Total width of the waveguide.
        :param co_resolution: Number of points along each axis in the grid on which the SDF is evaluated.
        :param sign: array, float, int or None determining the sign of the normal coordinates in parts of the cartesian
         coordinate system.
         None - sign is calculated automatically,
         float - the value is used a threshold for automatic calculation of the sign (same as None)
         int (+1 or -1) - the same sign everywhere
         array - array containing values of +1 or -1, specifying the sign for each point in the point cloud.
        """
        self._parameters = parameters
        VectorField.__init__(self, lcwg1_2d,  parameters, co_resolution, sign)

    @property
    def parameters(self):
        return self._parameters


class LCWG3Dm1(VectorField):

    def __init__(self,  parameters, co_resolution, sign):
        """
        M1 type vector field of a Liquid Crystal waveguide. Winding number in the yz plane is -1.
        The vector field is generated based on the distance field and size parameters of the waveguide.
        Point cloud specifying the values of the Signed Distance Function is taken as the input.
        The components of the output vector field are cartesian (x, y, z).
        :param parameters: Total width of the waveguide, thickness of the waveguide.
        :param co_resolution: Number of points along each axis in the grid on which the SDF is evaluated.
        :param sign: array, float, int or None determining the sign of the normal coordinates in parts of the cartesian
         coordinate system.
         None - sign is calculated automatically,
         float - the value is used a threshold for automatic calculation of the sign (same as None)
         int (+1 or -1) - the same sign everywhere
         array - array containing values of +1 or -1, specifying the sign for each point in the point cloud.
        """
        self._parameters = parameters
        VectorField.__init__(self, lcwg1_m1, parameters, co_resolution, sign)

    @property
    def parameters(self):
        return self._parameters


class LCWG3Dp1(VectorField):

    def __init__(self,  parameters, co_resolution, sign):
        """
        P1 type vector field of a Liquid Crystal waveguide. Winding number in the yz plane is +1.
        The vector field is generated based on the distance field and size parameters of the waveguide.
        Point cloud specifying the values of the Signed Distance Function is taken as the input.
        The components of the output vector field are cartesian (x, y, z).
        :param parameters: Total width of the waveguide, thickness of the waveguide.
        :param co_resolution: Number of points along each axis in the grid on which the SDF is evaluated.
        :param sign: array, float, int or None determining the sign of the normal coordinates in parts of the cartesian
         coordinate system.
         None - sign is calculated automatically,
         float - the value is used a threshold for automatic calculation of the sign (same as None)
         int (+1 or -1) - the same sign everywhere
         array - array containing values of +1 or -1, specifying the sign for each point in the point cloud.
        """
        self._parameters = parameters
        VectorField.__init__(self, lcwg1_p1, parameters, co_resolution, sign)

    @property
    def parameters(self):
        return self._parameters




# Copyright (C) 2024 Peter Ropač
# This file is part of SPOMSO.
# SPOMSO is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# SPOMSO is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public License along with SPOMSO. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from spomso.cores.vector_modification_functions import batch_normalize
from spomso.cores.helper_functions import smarter_reshape

# ----------------------------------------------------------------------------------------------------------------------
# VECTOR INITIALIZATION FUNCTIONS


def cartesian_define(p):
    ux = p[0]
    uy = p[1]
    uz = p[2]

    return np.asarray((ux, uy, uz))


def spherical_define(p):
    r = p[0]
    phi = p[1]
    theta = p[2]

    u = r*np.cos(phi)*np.sin(theta)
    v = r*np.sin(phi)*np.sin(theta)
    w = r*np.cos(theta)

    return np.asarray((u, v, w))


def cylindrical_define(p):
    r = p[0]
    phi = p[1]
    z = p[2]

    u = r*np.cos(phi)
    v = r*np.sin(phi)

    return np.asarray((u, v, z))


def radial_vector_field_spherical(r, *p):
    r = r.copy()
    return batch_normalize(r)


def radial_vector_field_cylindrical(r, *p):
    r = r.copy()
    r[2] = 0
    v = batch_normalize(r)
    return v


def hyperbolic_vector_field_cylindrical(r, *p):
    alpha = -np.arctan2(r[1], r[0])
    v = cylindrical_define(1, alpha, np.zeros(r.shape[1]))
    return v


def awn_vector_field_cylindrical(r, gamma):
    alpha = gamma*np.arctan2(r[1], r[0])
    print(alpha.shape, r.shape)
    v = cylindrical_define(1, np.squeeze(alpha), np.zeros(r.shape[1]))
    return v


def vortex_vector_field_cylindrical(r, *p):
    r = r.copy()
    r[2] = 0
    v = batch_normalize(r)

    w = v.copy()
    w[0] = -v[1]
    w[1] = v[0]
    return w


def aar_vector_field_cylindrical(r, alpha):
    r = r.copy()
    r[2] = 0
    vec = batch_normalize(r)
    print(alpha)
    sa = np.squeeze(np.sin(alpha))
    ca = np.squeeze(np.cos(alpha))

    o = np.asarray([vec[0, :] * ca - vec[1, :] * sa,
                    vec[0, :] * sa + vec[1, :] * ca,
                    vec[2, :]])

    return o


def aav_vector_field_cylindrical(r, alpha):
    r = r.copy()
    r[2] = 0
    vec = batch_normalize(r)
    print(alpha)
    sa = np.squeeze(np.sin(alpha))
    ca = np.squeeze(np.cos(alpha))

    o = np.asarray([-vec[0, :] * sa - vec[1, :] * ca,
                    vec[0, :] * ca - vec[1, :] * sa,
                    vec[2, :]])

    return o


def x_vector_field(r, *p):
    vec = np.zeros(r.shape)
    vec[0] = 1
    return vec


def y_vector_field(r, *p):
    vec = np.zeros(r.shape)
    vec[1] = 1
    return vec


def z_vector_field(r, *p):
    vec = np.zeros(r.shape)
    vec[2] = 1
    return vec


def from_sdf(sdf_, co_resolution):

    dimensions = np.asarray(co_resolution).shape[0]
    gsdf = smarter_reshape(sdf_, co_resolution)

    vec = np.asarray(np.gradient(gsdf))
    vec = vec.reshape(dimensions, -1)

    v = batch_normalize(vec)
    return v



# Copyright (C) 2024 Peter Ropač
# This file is part of SPOMSO.
# SPOMSO is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# SPOMSO is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public License along with SPOMSO. If not, see <https://www.gnu.org/licenses/>.

import numpy as np

from spomso.cores.transformations import EuclideanTransform, EuclideanTransformPoints
from spomso.cores.modifications import ModifyObject, ModifyVectorObject
from spomso.cores.helper_functions import resolution_conversion
from typing import Callable


class GenericGeometry(EuclideanTransform, ModifyObject):
    """Constructs geometry based on an SDF.

        Args:
            geo_sdf: SDF of a geometry - geo_sdf(co, *geo_parameters).
            geo_parameters: Parameters of the SDF.
    """

    def __init__(self, geo_sdf: Callable[[np.ndarray, tuple], np.ndarray], *geo_parameters: object):
        EuclideanTransform.__init__(self)
        ModifyObject.__init__(self, geo_sdf)
        self._geo_parameters = geo_parameters
        self._sdf = geo_sdf

    def create(self, co: np.ndarray) -> np.ndarray:
        """
        Applies the modifications and transformations (in that order) to the SDF and returns the map of the
        Signed Distance Field.

        Args:
            co: Point cloud of coordinates (D, N);
                D - number of dimensions (2,3);
                N - number of points in the point cloud.

        Returns:
            Signed Distance Field of shape (N,).
        """
        self._sdf = self.modified_object
        return self.apply(self._sdf, co, self.geo_parameters)

    def propagate(self, co: np.ndarray, *parameters_: tuple) -> np.ndarray:
        """
        Applies the modifications and transformations (in that order) to the SDF and returns the map of the
        Signed Distance Field. Similar to create but is meant to be used as input to construct new geometry.

        Args:
            co: Point cloud of coordinates (D, N);
                D - number of dimensions (2,3);
                N - number of points in the point cloud.
            parameters_: can be empty

        Returns:
            Signed Distance Field of shape (N,).
        """
        self._sdf = self.modified_object
        return self.apply(self._sdf, co, self.geo_parameters)

    def point_cloud(self, co: np.ndarray) -> np.ndarray:
        """Creates a point cloud from the interior points of the SDF.

        Args:
            co: Point cloud of coordinates (D, N);
        """
        data_ = self.create(co)
        mask_ = data_ <= 0

        _points = np.zeros((3, np.count_nonzero(mask_)))
        _points[:2, :] = co[:2, mask_]

        return _points


class Points(EuclideanTransformPoints):
    """Constructs a Points object from a point cloud.

    Args:
        points: The point cloud of shape (D, N), where D is the spatial dimension and N is the number of points.
    """

    def __init__(self, points: np.ndarray | list | tuple):
        EuclideanTransformPoints.__init__(self)
        self._points = np.asarray(points)
        if self._points.size > 0:
            if self._points.shape[1] < self._points.shape[0]:
                self._points = self._points.T

    @property
    def cloud(self) -> np.ndarray:
        """Transformed point cloud.

        Returns:
            Transformed point cloud of shape (D, N).
        """
        return self.apply(self._points)

    def from_image(self,
                   greyscale_image: np.ndarray,
                   image_dimensions: tuple | list | np.ndarray,
                   binary_threshold: float = 0.5):
        """
        Creates a point cloud from a greyscale image.
        Positions of the points in the point cloud are calculated
        from the positions of the pixels with brightness below the binary threshold.

        Args:
            greyscale_image: The greyscale image.
            image_dimensions: A 2vector defining the size of the image.
            binary_threshold: Threshold used to filter which points belong in the point cloud.
        """
        data_ = np.asarray(greyscale_image)
        data_max = np.amax(data_)
        data_ = data_.astype(float)/data_max

        data_ = np.flip(data_, axis=0).T
        mask_ = data_ <= binary_threshold

        coi = np.asarray(
            np.meshgrid(np.linspace(-image_dimensions[0] / 2, image_dimensions[0] / 2, data_.shape[0]),
                        np.linspace(-image_dimensions[1] / 2, image_dimensions[1] / 2, data_.shape[1]),
                        indexing="ij")
        )

        _points = np.zeros((3, np.count_nonzero(mask_)))
        _points[:2, :] = coi[:, mask_]

        self._points = _points

    def to_image(self, co_size: tuple | np.ndarray,
                 co_resolution: tuple | np.ndarray,
                 extend: tuple) -> np.ndarray:
        """
        Converts a point cloud into a black and white image.
        Black where there are points and white where there are none.

        Args:
            co_size: Size of the 3D coordinate system (size_x, size_y, size_z).
            co_resolution: Resolution of the grid along all three dimensions (res_x, res_y, res_z).
            extend: Along which directions is the image extended (-X, +X, -Y, +Y, -Z, +Z).

        Returns:
            A grid of voxels with the shape (res_x, res_y, res_z) and values of 0 or 1.
        """

        cloud = self.cloud

        new_res = (resolution_conversion(co_resolution[0]),
                   resolution_conversion(co_resolution[1]),
                   resolution_conversion(co_resolution[2]))

        out, edges = np.histogramdd(cloud.T,
                                    bins=new_res,
                                    range=((-co_size[0] / 2, co_size[0] / 2),
                                           (-co_size[1] / 2, co_size[1] / 2),
                                           (-co_size[2] / 2, co_size[2] / 2),)
                                    )
        out = out > 0
        out = out.astype(float)

        for ex in extend:

            if ex=="-Z" or ex=="+Z":
                p = np.amax(out, axis=(0,1))
                pp = p > 0
                zix = np.arange(0, new_res[2], dtype=int)
            if ex == "-Z":
                zminus_mask = zix < zix[pp][0]
                out[:, :, zminus_mask] = np.repeat(out[:, :, zix[pp][0]][:, :, np.newaxis],
                                                   np.count_nonzero(zminus_mask),
                                                   axis=2)
            if ex == "+Z":
                zplus_mask = zix > zix[pp][-1]
                out[:, :, zplus_mask] = np.repeat(out[:, :, zix[pp][-1]][:, :, np.newaxis],
                                                  np.count_nonzero(zplus_mask),
                                                  axis=2)

            if ex=="-X" or ex=="+X":
                p = np.amax(out, axis=(1, 2))
                pp = p > 0
                xix = np.arange(0, new_res[0], dtype=int)
            if ex == "-X":
                xminus_mask = xix < xix[pp][0]
                out[xminus_mask, :, :] = np.repeat(out[xix[pp][0], :, :][np.newaxis, :, :],
                                                   np.count_nonzero(xminus_mask),
                                                   axis=0)
            if ex == "+X":
                xplus_mask = xix > xix[pp][-1]
                out[xplus_mask, :, :] = np.repeat(out[xix[pp][-1], :, :][np.newaxis, :, :],
                                                  np.count_nonzero(xplus_mask),
                                                  axis=0)

            if ex=="-Y" or ex=="+Y":
                p = np.amax(out, axis=(0, 2))
                pp = p > 0
                yix = np.arange(0, new_res[1], dtype=int)
            if ex == "-Y":
                yminus_mask = yix < yix[pp][0]
                out[:, yminus_mask, :] = np.repeat(out[:, yix[pp][0], :][:, np.newaxis, :],
                                                   np.count_nonzero(yminus_mask),
                                                   axis=1)
            if ex == "+Y":
                yplus_mask = yix > yix[pp][-1]
                out[:, yplus_mask, :] = np.repeat(out[:, yix[pp][-1], :][:, np.newaxis, :],
                                                  np.count_nonzero(yplus_mask),
                                                  axis=1)

        return out


class VectorField(ModifyVectorObject):
    """Constructs a VectorField object from a vector field function.

        Args:
            vf: The vector field function - vf(p, *vf_parameters).
            vf_parameters: The parameters of the vector field function.
    """

    def __init__(self, vf: Callable[[np.ndarray, tuple], np.ndarray], *vf_parameters: tuple):
        ModifyVectorObject.__init__(self, vf)
        self._vf_parameters = vf_parameters
        self._vf = vf

    def create(self, p: np.ndarray) -> np.ndarray:
        """Applies the modifications to the vector field and returns the map of the vector field.
        
        Args:
            p: Point cloud of field components (D, N);
                D - number of dimensions (2,3);
                N - number of points in the point cloud.
        Returns:
             Vector field of shape (D, N).
        """
        self._vf = self.vf
        return self._vf(p, *self._vf_parameters)

    def propagate(self, p: np.ndarray, *parameters_: tuple) -> np.ndarray:
        """
        Applies the modifications to the vector field and returns the map of the vector field.
        Similar to create but is meant to be used as input to construct new vector fields.
        
        Args:
            p: Point cloud of field components (D, N);
                D - number of dimensions (2,3);
                N - number of points in the point cloud.
            parameters_: can be empty.
        
        Returns:
            Vector field of shape (D, N).
        """
        self._vf = self.vf
        return self._vf(p, *self._vf_parameters)

    def x(self, p: np.ndarray, *parameters_: tuple) -> np.ndarray:
        """
        Applies the modifications to the vector field and returns the x component of the vector field.
        Similar to create but is meant to be used as input to construct new vector fields.
        
        Args:
            p: Point cloud of field components (D, N);
                D - number of dimensions (2,3);
                N - number of points in the point cloud.
            parameters_: can be empty.
        
        Returns:
            Scalar field of shape (N,).
        """
        self._vf = self.vf
        return self._vf(p, *self._vf_parameters)[0]

    def y(self, p: np.ndarray, *parameters_: tuple) -> np.ndarray:
        """
        Applies the modifications to the vector field and returns the x component of the vector field.
        Similar to create but is meant to be used as input to construct new vector fields.
        
        Args:
            p: Point cloud of field components (D, N);
                D - number of dimensions (2,3);
                N - number of points in the point cloud.
            parameters_: can be empty.
        
        Returns:
            Scalar field of shape (N,).
        """
        self._vf = self.vf
        return self._vf(p, *self._vf_parameters)[1]

    def z(self, p: np.ndarray, *parameters_: tuple) -> np.ndarray:
        """
        Applies the modifications to the vector field and returns the x component of the vector field.
        Similar to create but is meant to be used as input to construct new vector fields.
        
        Args:
            p: Point cloud of field components (D, N);
                D - number of dimensions (2,3);
                N - number of points in the point cloud.
            parameters_: can be empty.
        
        Returns:
            Scalar field of shape (N,).
        """
        self._vf = self.vf
        return self._vf(p, *self._vf_parameters)[2]

    def phi(self, p: np.ndarray, *parameters_: tuple) -> np.ndarray:
        """
        Applies the modifications and returns the azimuthal angle for each vector in the vector field.
        Similar to create but is meant to be used as input to construct new vector fields.
        
        Args:
            p: Point cloud of field components (D, N);
                D - number of dimensions (2,3);
                N - number of points in the point cloud.
            parameters_: can be empty.
        
        Returns:
            Scalar field of shape (N,).
        """
        self._vf = self.vf
        vec = self._vf(p, *self._vf_parameters)

        return np.arctan2(vec[1], vec[0])

    def theta(self, p: np.ndarray, *parameters_: tuple) -> np.ndarray:
        """
        Applies the modifications and returns the polar angle for each vector in the vector field.
        Similar to create but is meant to be used as input to construct new vector fields.
        
        Args:
            p: Point cloud of field components (D, N);
                D - number of dimensions (2,3);
                N - number of points in the point cloud.
            parameters_: can be empty.
        
        Returns:
            Scalar field of shape (N,).
        """
        self._vf = self.vf
        vec = self._vf(p, *self._vf_parameters)

        return np.arccos(vec[2])

    def length(self, p: np.ndarray, *parameters_: tuple) -> np.ndarray:
        """
        Applies the modifications and returns the lengths of vectors in the vector field.
        Similar to create but is meant to be used as input to construct new vector fields.
        
        Args:
            p: Point cloud of field components (D, N);
                D - number of dimensions (2,3);
                N - number of points in the point cloud.
            parameters_: can be empty.
        
        Returns:
            Scalar field of shape (N,).
        """
        self._vf = self.vf
        vec = self._vf(p, *self._vf_parameters)

        return np.linalg.norm(vec, axis=0)








# Copyright (C) 2024 Peter Ropač
# This file is part of SPOMSO.
# SPOMSO is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# SPOMSO is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public License along with SPOMSO. If not, see <https://www.gnu.org/licenses/>.

from transformations import EuclideanTransform
from spomso.cores.modifications import ModifyObject
from sdf_3D import sdf_sphere, sdf_cylinder


class CylinderDirect(EuclideanTransform, ModifyObject):

    def __init__(self, radius, height):
        EuclideanTransform.__init__(self)
        ModifyObject.__init__(self, sdf_cylinder)
        self._radius = radius
        self._height = height
        self._sdf = sdf_cylinder

    def radius(self):
        return self._radius

    def height(self):
        return self._height

    def create(self, co):
        self._sdf = self.modified_object
        return self.apply(self._sdf, co, (self.radius, self.height))

    def propagate(self, co, *parameters_):
        self._sdf = self.modified_object
        return self.apply(self._sdf, co, (self.radius, self.height))


class SphereDirect(EuclideanTransform, ModifyObject):

    def __init__(self, radius):
        EuclideanTransform.__init__(self)
        ModifyObject.__init__(self, sdf_sphere)
        self._radius = radius
        self._sdf = sdf_sphere

    def radius(self):
        return self._radius

    def create(self, co):
        self._sdf = self.modified_object
        return self.apply(self._sdf, co, (self.radius,))

    def propagate(self, co, *parameters_):
        self._sdf = self.modified_object
        return self.apply(self._sdf, co, (self.radius,))



# Copyright (C) 2024 Peter Ropač
# This file is part of SPOMSO.
# SPOMSO is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# SPOMSO is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public License along with SPOMSO. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from spomso.cores.geom import GenericGeometry
from typing import Callable

def smoothmin_poly2(x, y, a):
    # https://iquilezles.org/articles/smin/
    h = np.maximum(a - np.abs(x - y), 0.0)/a
    return np.minimum(x, y) - h * h * a/4.0


def smoothmin_poly3(x, y, a):
    # https://iquilezles.org/articles/smin/
    if not a == 0.0:
        h = np.maximum(a - np.abs(x - y), 0.0)/a
        return np.minimum(x, y) - h * h * h * a / 6.0
    else:
        return np.minimum(y, y)


def smoothmax_boltz(x, y, width):
    # https: // en.wikipedia.org / wiki / Smooth_maximum
    exp1 = np.exp(x/width)
    exp2 = np.exp(y/width)

    return (x*exp1 + y*exp2)/(exp1 + exp2)


class CombineGeometry:
    """Class containing all the possible combination operations which can be applied to a scalar fields.

        Attributes:
            operations: Dictionary containing all the possible non-parametric operations.
            parametric_operations: Dictionary containing all the possible parametric operations.

        Args:
            operation_type: Type of operation with which two or more geometric objects are combined.
        """

    def __init__(self, operation_type: str):
        self.operation_type: str = operation_type
        self._combined_geometry: Callable[[np.ndarray, tuple], np.ndarray] = None
        self.operations: dict = {"UNION2": lambda obj1, obj2: np.minimum(obj1, obj2),
                                 "UNION": lambda *objs: np.amin(objs, axis=0),
                                 "SUBTRACT2": lambda obj1, obj2: np.maximum(obj1, -obj2),
                                 "INTERSECT2": lambda obj1, obj2: np.maximum(obj1, obj2),
                                 "INTERSECT": lambda *objs: np.amax(objs, axis=0),
                                 }

        self.parametric_operations: dict = {"SMOOTH_UNION2_2": lambda obj1, obj2, width: smoothmin_poly2(obj1,
                                                                                                         obj2,
                                                                                                         width),
                                            "SMOOTH_UNION2": lambda obj1, obj2, width: smoothmin_poly3(obj1,
                                                                                                       obj2,
                                                                                                       width),
                                            "SMOOTH_INTERSECT2": lambda obj1, obj2, width: -smoothmin_poly3(-obj1,
                                                                                                            -obj2,
                                                                                                            width),
                                            "SMOOTH_INTERSECT2_BOLTZMANN": lambda obj1, obj2, width: smoothmax_boltz(obj1,
                                                                                                                     obj2,
                                                                                                                     width),
                                            "SMOOTH_SUBTRACT2": lambda obj1, obj2, width: -smoothmin_poly3(-obj1,
                                                                                                           obj2,
                                                                                                           width),
                                            "SMOOTH_SUBTRACT2_BOLTZMANN": lambda obj1, obj2, width: smoothmax_boltz(obj1,
                                                                                                                    -obj2,
                                                                                                                    width)
                                      }

    @property
    def available_operations(self) -> list:
        """
        Available types of operations.
        
        Returns:
             List of available operations.
        """
        operations_list = list(self.operations.keys())
        print(f"Available non-parametric operations are: {operations_list}")
        return operations_list

    @property
    def available_parametric_operations(self) -> list:
        """
        Available types of operations which require a parameter.
        
        Returns:
              List of available parametric operations.
        """
        operations_list = list(self.parametric_operations.keys())
        print(f"Available parametric operations are: {operations_list}")
        return operations_list

    @property
    def combined_geometry(self) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Returns the SDF of the combined geometries,
        which can be used to create a new geometry using GenericGeometry class.
        
        Returns:
              SDF of the combined geometries
        """
        return self._combined_geometry

    def combine(self, *combined_objects: object) -> GenericGeometry:
        """
        Combines 2 or more geometric objects together.

        Args:
            combined_objects: Tuple containing geometric objects.
        
        Returns:
              New geometric object.
        """
        if self.operation_type not in self.operations.keys():
            raise SyntaxError(f"{self.operation_type} is not an implemented non-parametric operation.",
                              f"Possible operations are {self.operations.keys}")

        def new_geo_object(co, *params):
            sdfs = []
            for cobject in combined_objects:
                sdf_ = cobject.propagate(co, ())
                sdfs.append(sdf_)

            return self.operations[self.operation_type](*sdfs)

        self._combined_geometry = new_geo_object
        return GenericGeometry(new_geo_object, ())

    def combine_parametric(self, *combined_objects: object, parameters: tuple | float | int) -> GenericGeometry:
        """
        Combines 2 or more geometric objects together, based on the parameters of the operations.
        Args:
            combined_objects: Tuple containing geometric objects.
            parameters: Parameters of the operation.
        
        Returns:
              New geometric object.
        """
        if self.operation_type not in self.parametric_operations.keys():
            raise SyntaxError(f"{self.operation_type} is not an implemented parametric operation.",
                              f"Possible parametric operations are {self.parametric_operations.keys}")

        def new_geo_object(co, *params):
            sdfs = []
            for cobject in combined_objects:
                sdf_ = cobject.propagate(co, ())
                sdfs.append(sdf_)

            return self.parametric_operations[self.operation_type](*sdfs, parameters)

        self._combined_geometry = new_geo_object
        return GenericGeometry(new_geo_object, ())



# Copyright (C) 2024 Peter Ropač
# This file is part of SPOMSO.
# SPOMSO is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# SPOMSO is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public License along with SPOMSO. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from typing import Callable

# ----------------------------------------------------------------------------------------------------------------------
# VECTOR TRANSFORM FUNCTIONS


def batch_normalize(vec):

    m = np.linalg.norm(vec, axis=0)
    mask = ~(m == 0)
    vec[:, mask] = vec[:, mask]/m[mask]

    return vec


def add_vectors(vec, add_vec):
    add_vec = np.asarray(add_vec)
    if add_vec.size == 3:
        return np.add(vec.T, add_vec).T
    else:
        return np.add(vec, add_vec)


def subtract_vectors(vec, subtract_vec):
    subtract_vec = np.asarray(subtract_vec)
    if subtract_vec.size == 3:
        return np.add(vec.T, subtract_vec).T
    else:
        return np.add(vec, subtract_vec)


def rescale_vectors(vec, scale):

    return np.multiply(vec, scale)


def rotate_vectors_phi(vec, phis):
    sa = np.sin(phis)
    ca = np.cos(phis)

    o = np.asarray([vec[0, :] * ca - vec[1, :] * sa,
                    vec[0, :] * sa + vec[1, :] * ca,
                    vec[2, :]])

    return o


def rotate_vectors_theta(vec, thetas):

    rvec = vec.copy()
    rvec[2] = 0
    rvec = batch_normalize(rvec)

    ca = np.cos(thetas)
    sa = np.sin(thetas)

    term2 = np.asarray([rvec[0]*vec[2],
                        rvec[1]*vec[2],
                        -rvec[0]*vec[0] - rvec[1]*vec[1]])
    out = vec*ca + term2*sa
    return out


def rotate_vectors_x_axis(vec, alpha):
    sa = np.sin(alpha)
    ca = np.cos(alpha)

    o = np.asarray([vec[0, :],
                    vec[1, :] * ca - vec[2, :] * sa,
                    vec[1, :] * sa + vec[2, :] * ca,
                    ])

    return o


def rotate_vectors_y_axis(vec, alpha):
    sa = np.sin(alpha)
    ca = np.cos(alpha)

    o = np.asarray([vec[0, :] * ca - vec[2, :] * sa,
                    vec[1, :],
                    vec[0, :] * sa + vec[2, :] * ca,
                    ])

    return o


def rotate_vectors_z_axis(vec, alpha):
    sa = np.sin(alpha)
    ca = np.cos(alpha)

    o = np.asarray([vec[0, :] * ca - vec[1, :] * sa,
                    vec[0, :] * sa + vec[1, :] * ca,
                    vec[2, :]])

    return o


def rotate_vectors_axis(vec, axes, alpha):
    axes = np.asarray(axes)
    sa = np.sin(alpha)
    ca = np.cos(alpha)

    t1 = vec*ca
    t2 = sa*np.cross(axes.T, vec.T).T
    if axes.size==3:
        t3 = (1 - ca) * np.outer(axes, (np.sum(np.multiply(vec.T, axes.T).T, axis=0)))
    else:
        t3 = (1 - ca) * axes * (np.sum(np.multiply(vec.T, axes.T).T, axis=0))
    o = t1 + t2 + t3

    return o


def revolve_field_x(r, vec):

    alpha = np.arctan2(r[2], r[1])
    sa = np.sin(alpha)
    ca = np.cos(alpha)

    o = np.zeros(r.shape)
    o[0, :] = vec[0, :]
    o[1, :] = vec[1, :] * ca - vec[2, :] * sa
    o[2, :] = vec[1, :] * sa + vec[2, :] * ca

    return o


def revolve_field_y(r, vec):
    alpha = np.arctan2(r[2], r[0])
    sa = np.sin(alpha)
    ca = np.cos(alpha)

    o = np.zeros(r.shape)
    o[0, :] = vec[0, :] * ca - vec[2,:] * sa
    o[1, :] = vec[1, :]
    o[2, :] = vec[0, :] * sa + vec[2, :] * ca

    return o


def revolve_field_z(r, vec):
    alpha = np.arctan2(r[1], r[0])
    sa = np.sin(alpha)
    ca = np.cos(alpha)

    o = np.zeros(r.shape)
    o[0, :] = vec[0, :]*ca - vec[1, :]*sa
    o[1, :] = vec[0, :] * sa + vec[1, :] * ca
    o[2, :] = vec[2, :]

    return o




# Copyright (C) 2024 Peter Ropač
# This file is part of SPOMSO.
# SPOMSO is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# SPOMSO is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public License along with SPOMSO. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from scipy.interpolate import NearestNDInterpolator
from typing import Callable

from spomso.cores.vector_modification_functions import add_vectors, subtract_vectors, rescale_vectors
from spomso.cores.vector_modification_functions import rotate_vectors_phi, rotate_vectors_theta
from spomso.cores.vector_modification_functions import revolve_field_x, revolve_field_y, revolve_field_z
from spomso.cores.vector_modification_functions import rotate_vectors_x_axis
from spomso.cores.vector_modification_functions import rotate_vectors_y_axis
from spomso.cores.vector_modification_functions import rotate_vectors_z_axis
from spomso.cores.vector_modification_functions import rotate_vectors_axis
from spomso.cores.vector_modification_functions import batch_normalize

from spomso.cores.post_processing import sigmoid_falloff, positive_sigmoid_falloff, capped_exponential
from spomso.cores.post_processing import linear_falloff
from spomso.cores.post_processing import relu, smooth_relu, slowstart
from spomso.cores.post_processing import hard_binarization
from spomso.cores.post_processing import gaussian_boundary, gaussian_falloff
from spomso.cores.post_processing import conv_averaging, conv_edge_detection

from spomso.cores.helper_functions import smarter_reshape, vector_smarter_reshape


class ModifyObject:
    """Class containing all the possible modifications which can be applied to a scalar field.

    Attributes:
        original_geo_object: Original SDF.
        geo_object: Modified SDF.

    Args:
        geo_object: SDF of a geometry.
    """

    def __init__(self, geo_object: Callable[[np.ndarray, tuple], np.ndarray]):
        self._mod: list = []
        self.original_geo_object: Callable[[np.ndarray, tuple], np.ndarray] = geo_object
        self.geo_object: Callable[[np.ndarray, tuple], np.ndarray] = geo_object

    @property
    def modifications(self) -> list:
        """All the modifications which were applied to the geometry in chronological order.

        Returns:
            List of modifications.
        """
        return self._mod

    @property
    def modified_object(self) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        SDF of the modified geometry.

        Returns:
            SDF of the modified geometry.
        """
        return self.geo_object

    @property
    def original_object(self) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        SDF of the unmodified geometry.

        Returns:
            SDF of the unmodified geometry.
        """
        return self.original_geo_object

    def elongation(self, elon_vector: np.ndarray | tuple | list) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Elongates the geometry along a certain vector by the length of the vector in each respective direction.

        Args:
            elon_vector: 3vector defining the direction and distance of the elongation.
        
        Returns: 
            Modified SDF.
        """
        self._mod.append("elongation")
        elon_vector = np.asarray(elon_vector)/2

        geo_object = self.geo_object

        def new_geo_object(co, *params):
            qo0 = co[0] - np.clip(co[0], -elon_vector[0]/2, elon_vector[0]/2)
            qo1 = co[1] - np.clip(co[1], -elon_vector[1] / 2, elon_vector[1] / 2)
            qo2 = co[2] - np.clip(co[2], -elon_vector[2] / 2, elon_vector[2] / 2)
            qo = np.asarray([qo0, qo1, qo2])
            return geo_object(qo, *params)

        self.geo_object = new_geo_object
        return new_geo_object

    def rounding(self, rounding_radius: float | int) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Rounds off the geometry - effectively thickening it by the rounding radius.
        rounding_radius: Radius by which the edges are rounded and the object thickened.
        
        Returns: 
            Modified SDF.
        """
        self._mod.append("rounding")

        geo_object = self.geo_object

        def new_geo_object(co, *params):
            return geo_object(co, *params) - rounding_radius

        self.geo_object = new_geo_object
        return new_geo_object

    def rounding_cs(self,
                    rounding_radius: float | int,
                    bb_size: int | float) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Rounds off the geometry, but the geometry will be contained in its bounding box.
        Bounding box size must be specified.

        Args:
            rounding_radius: Radius by which the edges are rounded.
            bb_size: Size of the bounding box. Typically, the largest dimension of the object.
        
        Returns: 
            Modified SDF.
        """
        self._mod.append("rounding_cs")

        geo_object = self.geo_object

        def new_geo_object(co, *params):
            scale = (1 - rounding_radius/bb_size)
            return scale*geo_object(co/scale, *params) - rounding_radius

        self.geo_object = new_geo_object
        return new_geo_object

    def boundary(self) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Get the boundary of a shape.
        
        Returns: 
            Modified SDF.
        """
        self._mod.append("boundary")

        geo_object = self.geo_object

        def new_geo_object(co, *params):
            return np.abs(geo_object(co, *params))

        self.geo_object = new_geo_object
        return new_geo_object

    def signed_old(self, co_resolution: tuple) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Transform an Unsigned Distance Function into a Signed Distance Function.

        Args:
            co_resolution: Resolution of the coordinate system.
        
        Returns: 
            Modified SDF.
        """
        self._mod.append("signed")

        geo_object = self.geo_object

        def new_geo_object(co, *params):

            sp = geo_object(co, *params)
            if np.amin(sp) < 0:
                return sp

            s = smarter_reshape(sp, co_resolution)
            c = vector_smarter_reshape(co, co_resolution)

            seps = tuple(np.abs(c[i, 1*(i==0), 1*(i==1), 1*(i==2)] - c[i, 0, 0, 0]) for i in range(c.shape[0]))
            boundary = s < np.min(seps)

            mark = np.zeros(s.shape)
            fmark = np.zeros(s.shape)
            for i in range(1, s.shape[0]):
                chu = boundary[i, :, :] * ~boundary[i - 1, :, :]
                j = s.shape[0] - 1 - i
                chuu = boundary[j, :, :] * ~boundary[j + 1, :, :]
                mark[i, :, :] = chu + mark[i-1, :, :]
                fmark[i, :, :] = chuu + fmark[i - 1, :, :]
            fmark = np.flip(fmark, axis=0)
            interior = np.clip((mark % 2 + fmark % 2), 0, 1)

            mark = np.zeros(s.shape)
            fmark = np.zeros(s.shape)
            for i in range(1, s.shape[1]):
                chu = boundary[:, i, :] * ~boundary[:, i - 1, :]
                j = s.shape[1]-1-i
                chuu = boundary[:, j, :] * ~boundary[:, j + 1, :]
                mark[:, i, :] = chu + mark[:, i - 1, :]
                fmark[:, i, :] = chuu + fmark[:, i - 1, :]
            fmark = np.flip(fmark, axis=1)
            interior *= np.clip((mark % 2 + fmark % 2), 0, 1)

            interior = conv_averaging(interior, (2, 2, 1), 1)
            sign = 1 - 2*(interior > 0.5)
            sign = sign.flatten()

            return sp*sign

        self.geo_object = new_geo_object
        return new_geo_object

    def signed(self, co_resolution: tuple) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Transform an Unsigned Distance Function into a Signed Distance Function.

        Args:
            co_resolution: Resolution of the coordinate system.
        
        Returns: 
            Modified SDF.
        """
        self._mod.append("signed")

        geo_object = self.geo_object

        def new_geo_object(co, *params):

            sp = geo_object(co, *params)
            if np.amin(sp) < 0:
                return sp

            s = smarter_reshape(sp, co_resolution)
            c = vector_smarter_reshape(co, co_resolution)

            seps = tuple(np.abs(c[i, 1*(i==0), 1*(i==1), 1*(i==2)] - c[i, 0, 0, 0]) for i in range(c.shape[0]))
            boundary = s < np.min(seps)

            chu = np.zeros(s.shape)
            chuu = np.zeros(s.shape)
            chu[1:] = boundary[1:, :, :] * ~boundary[:-1, :, :]
            chuu[:-1] = boundary[:-1, :, :] * ~boundary[1:, :, :]
            mark = np.cumsum(chu, axis=0)
            fmark = np.cumsum(chuu, axis=0)
            fmark = np.flip(fmark, axis=0)
            interior = np.clip((mark % 2 + fmark % 2), 0, 1)

            chu = np.zeros(s.shape)
            chuu = np.zeros(s.shape)
            chu[:, 1:, :] = boundary[:, 1:, :] * ~boundary[:, :-1, :]
            chuu[:, :-1, :] = boundary[:, :-1, :] * ~boundary[:, 1:, :]
            mark = np.cumsum(chu, axis=1)
            fmark = np.cumsum(chuu, axis=1)
            fmark = np.flip(fmark, axis=1)
            interior *= np.clip((mark % 2 + fmark % 2), 0, 1)

            interior = conv_averaging(interior, (2, 2, 1), 1)

            interior = interior[1:-1, 1:-1, 1:-1]
            interior = np.pad(interior, pad_width=1, mode="edge")

            sign = 1 - 2*(interior > 0.5)
            sign = sign.flatten()

            return sp*sign

        self.geo_object = new_geo_object
        return new_geo_object

    def invert(self, direct: bool = False) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Invert the sign of the SDF.

        Args:
            direct: If True a function which returns inverts the sign of the SDF is returned without modifying the SDF.
                If False the same happens as described above but the SDF is also modified.
        
        Returns: 
            Modified SDF.
        """
        self._mod.append("invert")

        geo_object = self.geo_object

        def new_geo_object(co, *params):
            return -geo_object(co, *params)

        if not direct:
            self.geo_object = new_geo_object
        return new_geo_object

    def sign(self, direct: bool = False) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Get the sign of the SDF.

        Args:
            direct: If True a function which returns the sign of the SDF is returned without modifying the SDF.
                If False the same happens as described above but the SDF is also modified.
        
        Returns: 
            Modified SDF.
        """
        self._mod.append("sign")

        geo_object = self.geo_object

        def new_geo_object(co, *params):
            return np.sign(geo_object(co, *params))

        if not direct:
            self.geo_object = new_geo_object

        return new_geo_object

    def recover_volume(self,
                      interior: Callable[[np.ndarray, tuple], np.ndarray]) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Recovers the interior of the SDF if a function which outputs the correct value (-1 or 1) is provided.
        This function should take the same parameters as the SDF.
        Typically, this function is the output of self.sign().

        Args:
            interior: Function defining the interior (negative values) of an SDF.
        
        Returns: 
            Modified SDF.
        """
        self._mod.append("recover_volume")

        geo_object = self.geo_object

        def new_geo_object(co, *params):
            return geo_object(co, *params)*interior(co, *params)

        self.geo_object = new_geo_object
        return new_geo_object

    def define_volume(self,
                      interior: Callable[[np.ndarray, tuple], np.ndarray],
                      interior_parameters: tuple) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Defines the interior of the SDF if with a function.

        Args:
            interior: Function defining the interior (-1) and exterior (1) of an SDF.
            interior_parameters: Parameters of the function defining the interior of the SDF.
        
        Returns: 
            Modified SDF.
        """
        self._mod.append("define_volume")

        geo_object = self.geo_object

        def new_geo_object(co, *params):
            return geo_object(co, *params)*interior(co, *interior_parameters)

        self.geo_object = new_geo_object
        return new_geo_object

    def onion(self, thickness: float | int) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Transforms the geometry into a surface with some thickness.
        Args:
            thickness: Thickness of the resulting shape.
        
        Returns: 
            Modified SDF.
        """
        self._mod.append("onion")

        geo_object = self.geo_object

        def new_geo_object(co, *params):
            return np.abs(geo_object(co, *params)) - thickness

        self.geo_object = new_geo_object
        return new_geo_object

    def concentric(self, width: float | int) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Transforms an isosurface into two concentric isosurfaces which are apart by the value of width.
        Transforms a volume into an isosurface and rounds it by width/2.

        Args:
            width: Thickness of the resulting shape.
        
        Returns: 
            Modified SDF.
        """
        self._mod.append("concentric")

        geo_object = self.geo_object

        def new_geo_object(co, *params):
            return np.abs(geo_object(co, *params) - width/2)

        self.geo_object = new_geo_object
        return new_geo_object

    def revolution(self, radius: float | int) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Revolves a 2D shape around the y-axis to generate a 3D shape.
        First the 2D shape is translated along the x-axis by the radius of revolution,
        then it is revolved around the y-axis.

        Args:
            radius: Radius of revolution.
        
        Returns: 
            Modified SDF.
        """
        self._mod.append("revolution")

        geo_object = self.geo_object

        def new_geo_object(co, *params):
            mag_xz = np.linalg.norm([co[0,:], co[2,:]], axis=0)
            qo = np.zeros(co.shape)
            qo[0] = mag_xz - radius
            qo[1] = co[1]

            return geo_object(qo, *params)

        self.geo_object = new_geo_object
        return new_geo_object

    def axis_revolution(self, radius: float | int, angle: float | int) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Revolves a 2D shape around an axis to generate a 3D shape.
        First the 2D shape is translated along the x-axis by the radius of revolution,
        then it is revolved around the axis of revolution.
        The axis of revolution is angled by the specified angle with respect to the y-axis.

        Args:
            radius: Radius of revolution.
            angle: Angle between the axis of revolution and the y-axis.
        
        Returns: 
            Modified SDF.
        """
        self._mod.append("revolution")

        geo_object = self.geo_object

        def new_geo_object(co, *params):

            rot = np.asarray([[np.cos(angle), np.sin(angle)],[-np.sin(angle), np.cos(angle)]])
            co[:2, :] = rot.dot(co[:2])

            mag_xz = np.linalg.norm([co[0, :], co[2, :]], axis=0)

            qo = np.zeros(co.shape)
            qo[0] = mag_xz
            qo[1] = co[1]
            qo[:2] = rot.T.dot(qo[:2])
            qo[0] -= radius

            return geo_object(qo, *params)

        self.geo_object = new_geo_object
        return new_geo_object

    def extrusion(self, distance: float | int) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Extrudes a 2D shape along the z-axis so that the height of the object is equal to the specified distance.

        Args:
            distance: Final height of the object - distance of extrusion.
        
        Returns: 
            Modified SDF.
        """
        self._mod.append("extrusion")

        geo_object = self.geo_object

        def new_geo_object(co, *params):
            d = geo_object(co[:2], *params)
            w = np.asarray((d, np.abs(co[2]) - distance/2))

            first_term = np.minimum(np.maximum(w[0], w[1]), 0)
            second_term = np.linalg.norm(np.maximum(w, 0), axis=0)

            return first_term + second_term

        self.geo_object = new_geo_object
        return new_geo_object

    def twist(self, pitch: float | int) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Twists the geometry around the z-axis based on the position along the z-axis and the pitch.

        Args:
            pitch: rad/unit length.
        
        Returns: 
            Modified SDF.
        """
        self._mod.append("twist")

        geo_object = self.geo_object

        def new_geo_object(co, *params):

            c = np.cos(pitch * co[2])
            s = np.sin(pitch * co[2])
            rot = np.asarray([[c,s],[-s,c]])
            qo = co.copy()
            qo[:2,:] = rot[0, :, :] * co[0, :] + rot[1, :, :] * co[1, :]

            return geo_object(qo, *params)

        self.geo_object = new_geo_object
        return new_geo_object

    def bend2(self, radius, angle):
        # NOT WORKING!!!!!!
        self._mod.append("bend")

        geo_object = self.geo_object

        def new_geo_object(co, *params):


            f = np.clip(co[0]/radius, -angle/2, angle/2)
            c = np.cos(f)
            s = np.sin(f)

            qo = co.copy()
            qo[0] = s*radius - co[0] -s*co[1]
            qo[1] = radius*(1-c) + co[1]*(c-1)
            qo[2] = 0

            bo = co.copy()
            absco = np.abs(co[0])
            sco = np.sign(co[0])
            bo[2] = 0
            bo[0] = sco*(absco - radius*angle/2)*np.cos(angle/2)
            bo[1] = sco*bo[0] * np.tan(angle / 2)

            wo = qo + co + bo*(absco>radius*angle/2)

            # qo[:2, :] = rot[0, :, :] * co[0, :] + rot[1, :, :] * co[1, :]

            return geo_object(co, *params)

        self.geo_object = new_geo_object
        return new_geo_object

    def bend(self, radius: float | int, angle: float | int) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Bends the geometry around the z-axis, based on the specified bending radius and angle.
        The length of the bent section is radius*angle.

        Args:
            radius: Bending radius.
            angle: Bending angle.
        
        Returns: 
            Modified SDF.
        """
        self._mod.append("bend")

        geo_object = self.geo_object

        def new_geo_object(co, *params):

            c = np.cos(angle/2)
            s = np.sin(angle/2)
            rot = np.asarray([[c,s],[-s,c]])

            qo = co.copy()
            qo[1] -= radius
            phi = np.arctan2(qo[0], -qo[1])
            qo[1] = -radius + np.linalg.norm(qo[:2], axis=0)
            qo[0] = radius * phi

            mask1 = radius * angle / 2 <= np.abs(qo[0])
            mask2 = co[0, mask1] >= 0

            wo = co[:2].copy()
            wo = wo[:, mask1]

            sign = np.sign(co[0, mask1])

            wo[0] -= radius*s*sign
            wo[1] -= radius*(1-c)

            wo[:2, mask2] = rot.dot(wo[:2, mask2])
            wo[:2, ~mask2] = rot.T.dot(wo[:2, ~mask2])

            wo[0] += radius * (angle/2) * sign

            qo[:2, mask1] = wo
            return geo_object(qo, *params)

        self.geo_object = new_geo_object
        return new_geo_object

    def displacement(self,
                     displacement_function: Callable[[np.ndarray, tuple], np.ndarray],
                     displacement_function_parameters: tuple) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Displaces the surface of an object based on the values of the displacement function. Should be applied last.

        Args:
            displacement_function:
                Function which takes an array of coordinates with shape (D, N) and a tuple of parameters.
                D - number of dimensions (2 or 3);
                N - number of points in the point cloud.
            displacement_function_parameters: Parameters of the displacement function.
        
        Returns: 
            Modified SDF.
        """
        self._mod.append("displacement")

        geo_object = self.geo_object

        def new_geo_object(co, *params):

            return geo_object(co, *params) + displacement_function(co, *displacement_function_parameters)

        self.geo_object = new_geo_object
        return new_geo_object

    def infinite_repetition(self, distances: tuple | list | np.ndarray) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Infinitely repeats geometry in space on a cubic lattice.
        
        Args:
            distances: 3vector determining the distances between instances along each axis.
        
        Returns: 
            Modified SDF.
        """
        self._mod.append("infinite_repetition")

        geo_object = self.geo_object

        def new_geo_object(co, *params):

            redistances = np.asarray(distances)/2
            qo = np.mod(np.add(co.T, redistances), distances) - redistances

            return geo_object(qo.T, *params)

        self.geo_object = new_geo_object
        return new_geo_object

    def finite_repetition(self,
                          size: tuple | list | np.ndarray,
                          repetitions: tuple | list | np.ndarray) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Repeats the geometry finite amount of times along each axis within a defined bounding box.
        
        Args:
            size: Size of the bounding box along each axis in which the geometry is repeated.
            repetitions: Number of repetitions along each axis inside the bounding box.
        
        Returns: 
            Modified SDF.
        """
        self._mod.append("finite_repetition")

        size = np.asarray(size)
        rep = np.asarray(repetitions)

        geo_object = self.geo_object

        def new_geo_object(co, *params):

            c = size*(1 - 1/rep)/2
            d = size*(1/2 - 1/rep)
            s = size/rep
            mx = (co[0, :] >= -d[0])*(co[0, :] <= d[0])
            my = (co[1, :] >= -d[1])*(co[1, :] <= d[1])
            mz = (co[2, :] >= -d[2])*(co[2, :] <= d[2])

            # outer
            v = np.abs(co)
            v = np.subtract(v.T, c)
            v[:, 0] -= 2 * v[:, 0] * (co[0, :] < 0)
            v[:, 1] -= 2 * v[:, 1] * (co[1, :] < 0)
            v[:, 2] -= 2 * v[:, 2] * (co[2, :] < 0)

            # inner
            u = np.subtract(co.T, d)
            u[:] = np.mod(u[:], s) - s/2
            v[mx, 0] = u[mx,0]
            v[my, 1] = u[my, 1]
            v[mz, 2] = u[mz, 2]

            return geo_object(v.T, *params)

        self.geo_object = new_geo_object
        return new_geo_object

    def finite_repetition_rescaled(self,
                                   size: tuple | list | np.ndarray,
                                   repetitions: tuple | list | np.ndarray,
                                   instance_size: tuple | list | np.ndarray,
                                   padding: tuple | list | np.ndarray) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Repeats the geometry finite amount of times along each axis within a defined bounding box.
        The geometry is rescaled based on the provided geometry bounding box size and padding along each axis.
        Args:
            size: Size of the bounding box along each axis in which the geometry is repeated.
            repetitions: Number of repetitions along each axis inside the bounding box.
            instance_size: Size of the bounding box around one instance of geometry along each axis.
            padding: Padding around each instance along each axis.
        
        Returns: 
            Modified SDF.
        """
        self._mod.append("finite_repetition_rescaled")

        size = np.asarray(size)
        padding = np.asarray(padding)
        rep = np.asarray(repetitions)
        f = np.asarray(instance_size)

        geo_object = self.geo_object

        def new_geo_object(co, *params):

            c = size*(1 - 1/rep)/2
            d = size*(1/2 - 1/rep)
            s = size/rep
            mx = (co[0, :] >= -d[0])*(co[0, :] <= d[0])
            my = (co[1, :] >= -d[1])*(co[1, :] <= d[1])
            mz = (co[2, :] >= -d[2])*(co[2, :] <= d[2])

            # outer
            v = np.abs(co)
            v = np.subtract(v.T, c)
            v[:, 0] -= 2 * v[:, 0] * (co[0, :] < 0)
            v[:, 1] -= 2 * v[:, 1] * (co[1, :] < 0)
            v[:, 2] -= 2 * v[:, 2] * (co[2, :] < 0)

            # inner
            u = np.subtract(co.T, d)
            u[:] = np.mod(u[:], s) - s/2
            v[mx, 0] = u[mx,0]
            v[my, 1] = u[my, 1]
            v[mz, 2] = u[mz, 2]

            sss = np.min(s/(f + padding))
            w = v/sss

            return sss*geo_object(w.T, *params)

        self.geo_object = new_geo_object
        return new_geo_object

    def symmetry(self, axis: int) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Applies symmetry along an axis of the object.

        Args:
            axis: Index of the axis along which the symmetry is applied.
        
        Returns: 
            Modified SDF.
        """
        self._mod.append("symmetry")

        geo_object = self.geo_object

        def new_geo_object(co, *params):

            if axis > co.shape[0]:
                return geo_object(co, *params)

            co[axis, :] = np.abs(co[axis, :])
            return geo_object(co, *params)

        self.geo_object = new_geo_object
        return new_geo_object

    def mirror(self,
               a: tuple | list | np.ndarray,
               b: tuple | list | np.ndarray) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Applies mirroring along an axis connecting the two specified points.

        Args:
            a: Position of the object's mirror image.
            b: Position of the object.
        
        Returns: 
            Modified SDF.
        """
        self._mod.append("mirror")

        a = np.asarray(a)
        b = np.asarray(b)
        geo_object = self.geo_object

        def new_geo_object(co, *params):
            w = b - a
            c = (b + a) / 2
            l = np.linalg.norm(w)

            x = w / l
            y = np.asarray([-x[1], x[0], 0])
            y = y / np.linalg.norm(y)
            z = np.cross(x, y)
            rot = np.asarray([x, y, z])
            co = np.subtract(co.T, c).T
            co = rot.dot(co)

            # outer
            v = np.zeros(co.shape)
            v[1:, :] = co[1:]
            v[0, :] = np.abs(co[0, :]) - l / 2

            return geo_object(v, *params)

        self.geo_object = new_geo_object
        return new_geo_object

    def rotational_symmetry(self,
                            n: int,
                            radius: float | int,
                            phase: float | int) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Applies n-fold rotational symmetry to the object.

        Args:
            n: Order of the symmetry.
            radius: By how much the geometry is moved from the centre.
            phase: Angle by which the circular pattern is rotated by.
        
        Returns: 
            Modified SDF.
        """
        self._mod.append("rotational_symmetry")

        angle = 2*np.pi/n
        geo_object = self.geo_object

        def new_geo_object(co, *params):
            rot = np.asarray([[np.cos(angle/2 - phase), np.sin(angle/2 - phase)],
                              [-np.sin(angle/2 - phase), np.cos(angle/2 - phase)]])
            co[:2] = rot.dot(co[:2])
            phi = np.arctan2(co[1], co[0])
            phi[phi < 0] = 2 * np.pi + phi[phi < 0]
            phi = np.mod(phi, angle) - angle/2
            radii = np.linalg.norm(co[:2,:], axis=0)
            co[0] = radii*np.cos(phi) - radius
            co[1] = radii*np.sin(phi)

            return geo_object(co, *params)

        self.geo_object = new_geo_object
        return new_geo_object

    def linear_instancing(self,
                          n: int,
                          a: tuple | list | np.ndarray,
                          b: tuple | list | np.ndarray) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Instances the geometry n-times along the length of a segment defined by two points.

        Args:
            n: Number of instances along the segment.
            a: Starting point of the segment.
            b: Ending point of the segment.
        
        Returns: 
            Modified SDF.
        """
        self._mod.append("linear_instancing")

        a = np.asarray(a)
        b = np.asarray(b)
        geo_object = self.geo_object

        def new_geo_object(co, *params):
            w = b-a
            c = (b + a) / 2
            l = np.linalg.norm(w)

            x = w/l
            y = np.asarray([-x[1], x[0], 0])
            y = y/np.linalg.norm(y)
            z = np.cross(x,y)
            rot = np.asarray([x,y,z])
            co = np.subtract(co.T, c).T
            co = rot.dot(co)

            s = l/(n-1)
            d = s/2
            mx = (co[0, :] >= -l/2 + d) * (co[0, :] <= l/2 - d)

            # outer
            v = np.zeros(co.shape)
            v[1:, :] = co[1:]
            v[0, :] = np.abs(co[0,:]) - l/2
            v[0, :] -= 2*v[0, :]*(co[0, :] < 0)

            # inner
            if n > 2:
                u = np.subtract(co.T, [l/2 - d, 0, 0]).T
                u[:] = np.mod(u[:], s) - d
                v[0, mx] = u[0, mx]

            return geo_object(v, *params)

        self.geo_object = new_geo_object
        return new_geo_object

    def curve_instancing(self,
                         f: Callable[[np.ndarray, tuple], np.ndarray],
                         f_parameters: tuple,
                         t_range: tuple) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Instances the geometry n-times along the length of a parametric curve defined by a function.

        Args:
            f: Function defining the parametric curve.
            f_parameters: Parameters of the function defining the parametric curve.
            t_range: (t_start, t_end, n_instances):
                t_start - initial value of the parameter t, which is feed into the parametric curve.
                t_end - final value of the parameter t.
                n_instances -  number of instances of the geometry along the parametric curve.
        
        Returns: 
            Modified SDF.
        """
        self._mod.append("parametric_curve_instancing")

        geo_object = self.geo_object

        def new_geo_object(co, *params):

            ts = np.linspace(*t_range)
            fval = f(ts, *f_parameters)
            va = np.zeros((3, t_range[-1]))
            va[:fval.shape[0]] = fval
            ix = np.arange(0, t_range[-1])

            interp = NearestNDInterpolator(va.T, ix)
            values = interp(co[0], co[1], co[2])

            w = np.zeros(co.shape)
            for i in range(t_range[-1]):
                mask = values == i
                v = np.subtract(co[:, mask].T, va[:, i]).T
                w[:, mask] = v[:, :]

            return geo_object(w, *params)

        self.geo_object = new_geo_object
        return new_geo_object

    def aligned_curve_instancing(self,
                         f: Callable[[np.ndarray, tuple], np.ndarray],
                         f_parameters: tuple,
                         t_range: tuple) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Instances the geometry n-times along the length of a parametric curve defined by a function.
        The x-axis of the geometry is aligned with the tangent vector of the parametric curve.

        Args:
            f: Function defining the parametric curve.
            f_parameters: Parameters of the function defining the parametric curve.
            t_range: (t_start, t_end, n_instances):
                t_start - initial value of the parameter t, which is feed into the parametric curve.
                t_end - final value of the parameter t.
                n_instances -  number of instances of the geometry along the parametric curve.
        
        Returns: 
            Modified SDF.
        """
        # approximation with the binormal (dy) and normal (dz) vector!!!
        self._mod.append("aligned_parametric_curve_instancing")

        geo_object = self.geo_object
        tol = 0.001

        def new_geo_object(co, *params):

            ts = np.linspace(*t_range)
            ts_min = ts-tol
            ts_max = ts+tol

            fval = f(ts, *f_parameters)
            va = np.zeros((3, t_range[-1]))
            va[:fval.shape[0]] = fval
            ix = np.arange(0, t_range[-1])

            fv_min = f(ts_min, *f_parameters)
            fv_max = f(ts_max, *f_parameters)
            der = (fv_max-fv_min)/(2*tol)
            der = der/np.linalg.norm(der, axis=0)

            dx = np.zeros((3, t_range[-1]))
            dx[:fval.shape[0]] = der
            dy = np.zeros((3, t_range[-1]))
            dy[0, :] = -dx[1]
            dy[1, :] = dx[0]
            dz = np.cross(dx.T, dy.T).T
            trot = np.asarray([dx, dy, dz])

            interp = NearestNDInterpolator(va.T, ix)
            values = interp(co[0], co[1], co[2])

            w = np.zeros(co.shape)
            for i in range(t_range[-1]):
                mask = values == i
                v = np.subtract(co[:, mask].T, va[:, i]).T
                v = trot[:, :, i].dot(v)
                w[:, mask] = v[:, :]

            return geo_object(w, *params)

        self.geo_object = new_geo_object
        return new_geo_object

    def fully_aligned_curve_instancing(self,
                         f: Callable[[np.ndarray, tuple], np.ndarray],
                         f_parameters: tuple,
                         t_range: tuple) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Instances the geometry n-times along the length of a parametric curve defined by a function.
        The x-axis of the geometry is aligned with the tangent vector of the parametric curve.
        The y-axis of the geometry is aligned with the normal vector of the parametric curve.
        The z-axis of the geometry is aligned with the binormal vector of the parametric curve.

        Args:
            f: Function defining the parametric curve.
            f_parameters: Parameters of the function defining the parametric curve.
            t_range: (t_start, t_end, n_instances):
                t_start - initial value of the parameter t, which is feed into the parametric curve.
                t_end - final value of the parameter t.
                n_instances -  number of instances of the geometry along the parametric curve.
        
        Returns: 
            Modified SDF.
        """
        self._mod.append("fully_aligned_parametric_curve_instancing")

        geo_object = self.geo_object
        tol = 0.001

        def new_geo_object(co, *params):

            ts = np.linspace(*t_range)
            ts_min = ts-tol
            ts_max = ts+tol

            fval = f(ts, *f_parameters)
            va = np.zeros((3, t_range[-1]))
            va[:fval.shape[0]] = fval
            ix = np.arange(0, t_range[-1])

            fv_min = f(ts_min, *f_parameters)
            fv_max = f(ts_max, *f_parameters)
            der = (fv_max-fv_min)/(2*tol)
            dermag = np.linalg.norm(der, axis=0)
            der = der/dermag

            der2 = (fv_max - 2*fval + fv_min)/(tol**2)
            der2 = der2/dermag
            der2mag = np.linalg.norm(der2, axis=0)
            der2 = der2/der2mag

            dx = np.zeros((3, t_range[-1]))
            dx[:fval.shape[0]] = der
            dy = np.zeros((3, t_range[-1]))
            dy[:fval.shape[0]] = der2
            dz = np.cross(dx.T, dy.T).T
            trot = np.asarray([dx, dy, dz])

            interp = NearestNDInterpolator(va.T, ix)
            values = interp(co[0], co[1], co[2])

            w = np.zeros(co.shape)
            for i in range(t_range[-1]):
                mask = values == i
                v = np.subtract(co[:, mask].T, va[:, i]).T
                v = trot[:, :, i].dot(v)
                w[:, mask] = v[:, :]

            return geo_object(w, *params)

        self.geo_object = new_geo_object
        return new_geo_object

    def custom_modification(self,
                            modification: Callable[[Callable[[np.ndarray, tuple], np.ndarray],
                                                    np.ndarray,
                                                    tuple,
                                                    tuple],
                            np.ndarray],
                            modification_parameters: tuple,
                            modification_name: str = "custom") -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Applies a custom user-specified modification to a scalar (Signed Distance Function) field.

        Args:
            modification: A custom modification which takes the SDF as a first argument,
                point cloud of coordinates at which the SDF is evaluated as the second argument,
                parameters of the SDF as the third argument,
                and the custom modification parameters as the fourth argument.
            modification_parameters: Parameters of the custom modification.
            modification_name: Name of the custom modification.

        Returns:
            Modified Scalar Field.
        """
        self._mod.append(modification_name)
        geo_object = self.geo_object

        def new_geo_object(co, *params):
            return modification(geo_object, co, params, modification_parameters)

        self.geo_object = new_geo_object
        return new_geo_object

    def sigmoid_falloff(self, amplitude: float | int, width: float | int) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Applies a sigmoid to the scalar (Signed Distance Function) field.

        Args:
            amplitude: Maximum value of the transformed scalar field.
            width: Width of the sigmoid.

        Returns:
            Modified Scalar Field.
        """
        self._pmod.append("sigmoid_falloff")
        geo_object = self.geo_object

        def new_geo_object(co, *params):
            u = geo_object(co, *params)
            return sigmoid_falloff(u, amplitude, width)

        self.geo_object = new_geo_object
        return new_geo_object

    def positive_sigmoid_falloff(self,
                                 amplitude: float | int,
                                 width: float | int) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Applies a sigmoid, shifted to the positive velues by the value of the width parameter,
        to the scalar (Signed Distance Function) field.

        Args:
            amplitude: Maximum value of the transformed scalar field.
            width: Width of the sigmoid.

        Returns:
            Modified Scalar Field.
        """
        self._mod.append("positive_sigmoid_falloff")
        geo_object = self.geo_object

        def new_geo_object(co, *params):
            u = geo_object(co, *params)
            return positive_sigmoid_falloff(u, amplitude, width)

        self.geo_object = new_geo_object
        return new_geo_object

    def capped_exponential(self,
                           amplitude: float | int,
                           width: float | int) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Applies a decreasing exponential functon to the scalar (Signed Distance Function) field.
        to the scalar (Signed Distance Function) field.

        Args:
            amplitude: Maximum value of the transformed scalar field.
            width: Range at which the value of the transformed scalar field drops to almost zero.

        Returns:
            Modified Scalar Field.
        """
        self._mod.append("capped_exponential")
        geo_object = self.geo_object

        def new_geo_object(co, *params):
            u = geo_object(co, *params)
            return capped_exponential(u, amplitude, width)

        self.geo_object = new_geo_object
        return new_geo_object

    def hard_binarization(self, threshold: float) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Binarizes the Signed Distance field/pattern based on a threshold.
        Values below the threshold are 1 and values above are 0.

        Args:
            threshold: Binarization threshold.

        Returns:
            Modified Scalar Field.
        """
        self._mod.append("hard_binarization")
        geo_object = self.geo_object

        def new_geo_object(co, *params):
            u = geo_object(co, *params)
            return hard_binarization(u, threshold)

        self.geo_object = new_geo_object
        return new_geo_object

    def linear_falloff(self, amplitude: float | int, width: float | int) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Applies a decreasing linear function to the scalar (Signed Distance Function) field.

        Args:
            amplitude: Maximum value of the transformed scalar field.
            width: Range at which the value of the transformed scalar field drops to zero.

        Returns:
            Modified Scalar Field.
        """
        self._mod.append("linear_falloff")
        geo_object = self.geo_object

        def new_geo_object(co, *params):
            u = geo_object(co, *params)
            return linear_falloff(u, amplitude, width)

        self.geo_object = new_geo_object
        return new_geo_object

    def relu(self, width: float | int) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Applies the ReLU function to the scalar (Signed Distance Function) field.

        Args:
            width: Range at which the value of the transformed field reaches one.

        Returns:
            Modified Scalar Field.
        """
        self._mod.append("relu")
        geo_object = self.geo_object

        def new_geo_object(co, *params):
            u = geo_object(co, *params)
            return relu(u, width)

        self.geo_object = new_geo_object
        return new_geo_object

    def smooth_relu(self, smooth_width: float | int,
                    width: float | int = 1, threshold: int | float = 0.01) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Applies the "squareplus" function to the scalar (Signed Distance Function) field.
        https://en.wikipedia.org/wiki/Rectifier_(neural_networks)

        Args:
            smooth_width: Distance from the origin at which the Smooth ReLU function
                is greater than ReLU for less than the value of the threshold parameter.
            width: Range at which the value of the transformed field reaches one.
            threshold: At smooth_width distance from the origin the value of the Smooth ReLU function is greater
                than ReLU for the value of the threshold parameter.
                at smooth_width distance from the origin.

        Returns:
            Transformed scalar field.
        """
        self._mod.append("smooth_relu")
        geo_object = self.geo_object

        def new_geo_object(co, *params):
            u = geo_object(co, *params)
            return smooth_relu(u, smooth_width, width, threshold)

        self.geo_object = new_geo_object
        return new_geo_object

    def slowstart(self, smooth_width: float | int,
                  width: float | int = 1,
                  threshold: int | float = 0.01,
                  ground: bool = True) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Applies the SlowStart function to the scalar (Signed Distance Function) field.

        Args:
            smooth_width: Distance from the origin at which the SlowStart function
                is greater than ReLU for less than the value of the threshold parameter.
            width: Range at which the value of the transformed field reaches one.
            threshold: At smooth_width distance from the origin the value of the SlowStart function is greater
                than ReLU for the value of the threshold parameter.
            ground: if True the value of the function is zero at zero.

        Returns:
            Transformed scalar field.
        """

        self._mod.append("slowstart")
        geo_object = self.geo_object

        def new_geo_object(co, *params):
            u = geo_object(co, *params)
            return slowstart(u, smooth_width, width, threshold, ground=ground)

        self.geo_object = new_geo_object
        return new_geo_object

    def gaussian_boundary(self, amplitude: float | int, width: float | int) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Applies the Gaussian to the scalar (Signed Distance Function) field.

        Args:
            amplitude: Maximum value of the transformed scalar field.
            width: Range at which the value of the transformed scalar field drops to almost zero.

        Returns:
            Modified Scalar Field.
        """
        self._mod.append("gaussian_boundary")
        geo_object = self.geo_object

        def new_geo_object(co, *params):
            u = geo_object(co, *params)
            return gaussian_boundary(u, amplitude, width)

        self.geo_object = new_geo_object
        return new_geo_object

    def gaussian_falloff(self, amplitude: float | int, width: float | int) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Applies the Gaussian to the positive values of the scalar (Signed Distance Function) field.

        Args:
            amplitude: Maximum value of the transformed scalar field (and points at which the scalar field was < 0).
            width: Range at which the value of the transformed scalar field drops to almost zero.

        Returns:
            Modified Scalar Field.
        """
        self._mod.append("gaussian_falloff")
        geo_object = self.geo_object

        def new_geo_object(co, *params):
            u = geo_object(co, *params)
            return gaussian_falloff(u, amplitude, width)

        self.geo_object = new_geo_object
        return new_geo_object

    def conv_averaging(self,
                       kernel_size: int | tuple | list | np.ndarray,
                       iterations: int,
                       co_resolution: tuple) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Averages the field using an averaging convolutional kernel of the specified size.

        Args:
            kernel_size: Size of the averaging kernel. Must be an integer or a tuple/array of the
                same dimension as the scaler field.
            iterations: Number of times the convolutional averaging is applied to the input scalar field.
            co_resolution: Resolution of the coordinate system.

        Returns:
            Modified Scalar Field.
        """
        self._mod.append("conv_averaging")
        geo_object = self.geo_object

        def new_geo_object(co, *params):
            u = geo_object(co, *params)
            u = smarter_reshape(u, co_resolution)
            out = conv_averaging(u, kernel_size, iterations)
            return out.flatten()

        self.geo_object = new_geo_object
        return new_geo_object

    def conv_edge_detection(self,
                            co_resolution: tuple) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Edge detection based ona 3x3 convolutional kernel.

        Args:
            co_resolution: Resolution of the coordinate system.

        Returns:
            Modified Scalar Field.
        """
        self._mod.append("conv_edge_detection")
        geo_object = self.geo_object

        def new_geo_object(co, *params):
            u = geo_object(co, *params)
            u = smarter_reshape(u, co_resolution)
            return conv_edge_detection(u)

        self.geo_object = new_geo_object
        return new_geo_object

    def custom_post_process(self,
                            function: Callable[[np.ndarray, tuple], np.ndarray],
                            parameters: tuple,
                            post_process_name: str = "custom") -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Applies a custom user-specified post-processing function to a scalar (Signed Distance Function) field.

        Args:
            function: A custom post-processing function which takes the SDF as a first argument
                and the parameters of the function as the next arguments.
            parameters: Parameters of the custom post-processing function.
            post_process_name: Name of the custom post-processing function.

        Returns:
             Modified Scalar Field.
        """
        self._mod.append(post_process_name)
        geo_object = self.geo_object

        def new_geo_object(co, *params):
            u = geo_object(co, *params)
            return function(u, *parameters)

        self.geo_object = new_geo_object
        return new_geo_object


class ModifyVectorObject:
    """Class containing all the possible modifications which can be applied to a vector field.

        Attributes:
            original_vf: Original vector field.
            vf: Modified vector field.

        Args:
            vf: Vector field.
        """

    def __init__(self, vf: Callable[[np.ndarray, tuple], np.ndarray]):
        self._mod: list = []
        self.original_vf: Callable[[np.ndarray, tuple], np.ndarray] = vf
        self.vf: Callable[[np.ndarray, tuple], np.ndarray] = vf

    @property
    def modifications(self) -> list:
        """
        All the modifications which were applied to the vector field in chronological order.

        Returns:
            List of modifications.
        """
        return self._mod

    @property
    def modified_object(self) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Modified vector field.

        Returns:
            Modified vector field.
        """
        return self.vf

    @property
    def original_object(self) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Unmodified vector field.

        Returns:
            Unmodified vector field.
        """
        return self.original_vf

    def add(self, second_field: int | float | np.ndarray) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Adds a number, vector, or a vector field of the same shape to the vector field.

        Args:
            second_field: a number, vector, or a vector field to be added.

        Returns:
            Modified vector field.
        """
        self._mod.append("add")

        vf = self.vf

        def new_vf(p, *params):
            return add_vectors(vf(p, *params), second_field)

        self.vf = new_vf
        return new_vf

    def subtract(self, second_field: int | float | np.ndarray) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Subtracts a number, vector, or a vector field of the same shape to the vector field.

        Args:
            second_field: a number, vector, or a vector field to be subtracted.

        Returns:
            Modified vector field.
        """
        self._mod.append("subtract")

        vf = self.vf

        def new_vf(p, *params):
            return subtract_vectors(vf(p, *params), second_field)

        self.vf = new_vf
        return new_vf

    def rescale(self, second_field: int | float | np.ndarray) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Rescales the lengths of vectors in the vector field by a number, vector, or a vector field of the same shape.

        Args:
            second_field: a number, vector, or a vector field used for rescaling.

        Returns:
            Modified vector field.
        """
        self._mod.append("rescale")

        vf = self.vf

        def new_vf(p, *params):
            return rescale_vectors(vf(p, *params), second_field)

        self.vf = new_vf
        return new_vf

    def rotate_phi(self, phi: int | float | np.ndarray) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Rotates vectors in a 3D vector field by some azimuthal angle.

        Args:
            phi: Spatially dependent or independent angle by which to rotate the vectors in the xy-plane.
                                  
        Returns:
            Modified vector field.
        """
        self._mod.append("rotate_phi")

        vf = self.vf

        def new_vf(p, *params):
            return rotate_vectors_phi(vf(p, *params), phi)

        self.vf = new_vf
        return new_vf

    def rotate_theta(self, theta: int | float | np.ndarray) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Rotates vectors in a 3D vector field by some polar angle.

        Args:
            theta: Spatially dependent or independent angle by which to rotate the vectors in the polar direction.
                                  
        Returns:
            Modified vector field.
        """
        self._mod.append("rotate_theta")

        vf = self.vf

        def new_vf(p, *params):
            return rotate_vectors_theta(vf(p, *params), theta)

        self.vf = new_vf
        return new_vf

    def rotate_x(self, alpha: int | float | np.ndarray) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Rotates vectors in a 3D vector field by some angle.

        Args:
            alpha: Spatially dependent or independent angle by which to rotate the vectors around the x axis.
                                  
        Returns:
            Modified vector field.
        """
        self._mod.append("rotate_x")

        vf = self.vf

        def new_vf(p, *params):
            return rotate_vectors_x_axis(vf(p, *params), alpha)

        self.vf = new_vf
        return new_vf

    def rotate_y(self, alpha: int | float | np.ndarray) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Rotates vectors in a 3D vector field by some angle.

        Args:
            alpha: Spatially dependent or independent angle by which to rotate the vectors around the y axis.
                                  
        Returns:
            Modified vector field.
        """
        self._mod.append("rotate_y")

        vf = self.vf

        def new_vf(p, *params):
            return rotate_vectors_y_axis(vf(p, *params), alpha)

        self.vf = new_vf
        return new_vf

    def rotate_z(self, alpha: int | float | np.ndarray) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Rotates vectors in a 3D vector field by some angle.

        Args:
            alpha: Spatially dependent or independent angle by which to rotate the vectors around the z axis.
                                  
        Returns:
            Modified vector field.
        """
        self._mod.append("rotate_z")

        vf = self.vf

        def new_vf(p, *params):
            return rotate_vectors_z_axis(vf(p, *params), alpha)

        self.vf = new_vf
        return new_vf

    def rotate_axis(self,
                    axis: tuple | list | np.ndarray,
                    alpha: int | float | np.ndarray) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Rotates vectors in a 3D vector field by some angle around some axis.

        Args:
            axis: A single or an array exes around which the vectors in the vector field will be rotated.
            alpha: Spatially dependent or independent angle by which to rotate the vectors around the specified axis.
                                  
        Returns:
            Modified vector field.
        """
        self._mod.append("rotate_axis")

        vf = self.vf

        def new_vf(p, *params):
            return rotate_vectors_axis(vf(p, *params), axis, alpha)

        self.vf = new_vf
        return new_vf

    def revolution_x(self, co: np.ndarray) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Revolves a 2D vector field around the x-axis to generate a 3D vector field.

        Args:
            co: Coordinate system with shape (3, N).
                                  
        Returns:
            Modified vector field.
        """
        self._mod.append("revolution_x")

        vf = self.vf

        def new_vf(p, *params):

            return revolve_field_x(co, vf(p, *params))

        self.vf = new_vf
        return new_vf

    def revolution_y(self, co: np.ndarray) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Revolves a 2D vector field around the y-axis to generate a 3D vector field.

        Args:
            co: Coordinate system with shape (3, N).
                                  
        Returns:
            Modified vector field.
        """
        self._mod.append("revolution_y")

        vf = self.vf

        def new_vf(p, *params):
            return revolve_field_y(co, vf(p, *params))

        self.vf = new_vf
        return new_vf

    def revolution_z(self, co: np.ndarray) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Revolves a 2D vector field around the z-axis to generate a 3D vector field.

        Args:
            co: Coordinate system with shape (3, N).
                                  
        Returns:
            Modified vector field.
        """
        self._mod.append("revolution_z")

        vf = self.vf

        def new_vf(p, *params):
            return revolve_field_z(co, vf(p, *params))

        self.vf = new_vf
        return new_vf

    def normalize(self) -> Callable[[np.ndarray, tuple], np.ndarray]:
        """
        Normalizes the vector field.
                                  
        Returns:
            Modified vector field.
        """
        self._mod.append("normalize")

        vf = self.vf

        def new_vf(p, *params):
            return batch_normalize(vf(p, *params))

        self.vf = new_vf
        return new_vf



# Copyright (C) 2024 Peter Ropač
# This file is part of SPOMSO.
# SPOMSO is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# SPOMSO is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public License along with SPOMSO. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from spomso.cores.helper_functions import smarter_reshape
from spomso.cores.vector_modification_functions import batch_normalize
from spomso.cores.vector_modification_functions import rotate_vectors_axis
from spomso.cores.post_processing import conv_averaging


def compute_crossings_2d(sdf_grid: np.ndarray, thr: float = 0.06) -> np.ndarray:
    """
    Calculates where boundaries in the SDF and separates the regions by assigning the a +1 or -1 value.
    :param sdf_grid: Signed Distance field or any scalar field.
    :param thr: Threshold by which the regions are separated.
    :return: Modified scalar field.
    """
    cross = np.zeros(sdf_grid.shape)

    s = sdf_grid[:, :]
    min_ = np.amin(s)
    c1 = np.ones(sdf_grid.shape)

    c1[:, 1:] = np.isclose(s[:, 1:], min_, atol=thr)*(~np.isclose(s[:, :-1], min_, atol=thr))

    for j in range(1, sdf_grid.shape[1]):
        cross[:, j] = 1*c1[:, j] + cross[:, j-1]

    o = np.mod(cross,2).T
    o = conv_averaging(o, 5, 2) >= 0.5
    o = 2 * o.T - 1

    return o


def lcwg1_2d_old(r, uu, p):

    w = p

    o = np.zeros((3, uu.shape[0]))

    l1 = o.copy()
    l1[0, :] = 1

    qq = np.clip(uu, 0, 1)

    l4 = o.copy()
    l4[0, :] = -np.cos(np.pi * qq)
    l4[1, :] = -np.sin(np.pi * qq)
    l4[:] = l4 / np.linalg.norm(l4, axis=0)

    wg_mask = np.abs(r[1]) <= w / 2

    o[:] = l1
    o[:, wg_mask] = l4[:, wg_mask]

    return o


def lcwg1_p1_old(r, uu, p):

    w = p[0]
    d = p[1]

    alpha = np.arctan2(2 * r[2] / (d ** 2), 8 * r[1] / (w ** 2))
    sa = np.sin(alpha)
    ca = np.cos(alpha)

    o = np.zeros((3, uu.shape[0]))

    l1 = o.copy()
    l1[0, :] = 1

    qq = np.clip(uu, 0, 1)

    l2 = -np.sin(np.pi * qq)

    l4 = o.copy()
    l4[0, :] = -np.cos(np.pi * qq)
    l4[1, :] = l2 * ca
    l4[2, :] = l2 * sa
    l4[:] = l4 / np.linalg.norm(l4, axis=0)

    wg_mask = np.abs(r[1]) <= w / 2

    o[:] = l1
    o[:, wg_mask] = l4[:, wg_mask]

    return o


def lcwg1_m1_old(r, uu, p):

    w = p[0]
    d = p[1]

    alpha = -np.arctan2(2 * r[2] / (d ** 2), 8 * r[1] / (w ** 2))
    sa = np.sin(alpha)
    ca = np.cos(alpha)

    o = np.zeros((3, uu.shape[0]))

    l1 = o.copy()
    l1[0, :] = 1

    qq = np.clip(uu, 0, 1)

    l2 = -np.sin(np.pi * qq)

    l4 = o.copy()
    l4[0, :] = -np.cos(np.pi * qq)
    l4[1, :] = l2 * ca
    l4[2, :] = l2 * sa
    l4[:] = l4 / np.linalg.norm(l4, axis=0)

    wg_mask = np.abs(r[1]) <= w / 2

    o[:] = l1
    o[:, wg_mask] = l4[:, wg_mask]

    return o


def lcwg1_2d(uu, p, co_resolution, sign):

    w = p

    dimensions = np.asarray(co_resolution).shape[0]
    gsdf = smarter_reshape(uu, co_resolution)

    vec = np.asarray(np.gradient(gsdf))
    vec = vec.reshape(dimensions, -1)

    # avec = np.abs(vec)
    # sign = np.sign(np.sum(np.multiply(vec, avec), axis=0))

    if sign is None or isinstance(sign, float):
        if isinstance(sign, float):
            s = compute_crossings_2d(gsdf[:, :, 0], thr=abs(sign))
        else:
            s = compute_crossings_2d(gsdf[:, :, 0])
        sign = np.repeat(s[:, :, np.newaxis], co_resolution[2], axis=2).flatten()

    qq = np.clip(2 * uu / w, 0, 1)
    phis = sign*(qq*np.pi + np.pi/2)
    sap = np.sin(phis)
    cap = np.cos(phis)

    l4 = vec.copy()
    l4 = np.asarray([l4[0, :] * cap - l4[1, :] * sap,
                     l4[0, :] * sap + l4[1, :] * cap,
                     l4[2, :]])

    o = np.zeros((3, uu.shape[0]))
    o[:dimensions, :] = l4
    o = batch_normalize(o)

    return o


def lcwg1_p1(uuww, p, co_resolution, sign):

    uu = uuww[0]
    ww = uuww[1]

    w = p[0]
    d = p[1]

    dimensions = np.asarray(co_resolution).shape[0]

    pp = np.linalg.norm([2 * uu / w, ww / d], axis=0)
    gsdf = smarter_reshape(pp, co_resolution)

    vec = np.asarray(np.gradient(gsdf))
    vec = vec.reshape(dimensions, -1)
    vec = batch_normalize(vec)

    e1 = vec.copy()
    e1[0] = -vec[1]
    e1[1] = vec[0]
    e1[2] = 0
    e1 = batch_normalize(e1)
    e2 = np.cross(vec.T, e1.T).T

    # avec = np.abs(vec)
    # sign = np.sign(np.sum(np.multiply(vec, avec), axis=0))

    if sign is None or isinstance(sign, float):
        if isinstance(sign, float):
            s = compute_crossings_2d(gsdf[:, :, 0], thr=abs(sign))
        else:
            s = compute_crossings_2d(gsdf[:, :, 0])
        sign = np.repeat(s[:, :, np.newaxis], gsdf.shape[2], axis=2).flatten()

    qq = np.clip(pp, 0, 1)
    phis = sign * (+qq * np.pi + np.pi / 2)
    vec2 = rotate_vectors_axis(vec, e2, phis)

    o = np.zeros((3, uu.shape[0]))
    alpha = np.arctan2(2 * ww / (d ** 2), 8 * uu / (w ** 2))
    o[:dimensions, :] = rotate_vectors_axis(vec2, e1, -2 * alpha)
    o = batch_normalize(o)

    return o


def lcwg1_m1(uuww, p, co_resolution, sign):

    uu = uuww[0]
    ww = uuww[1]

    w = p[0]
    d = p[1]

    dimensions = np.asarray(co_resolution).shape[0]

    pp = np.linalg.norm([2 * uu / w, ww / d], axis=0)
    gsdf = smarter_reshape(pp, co_resolution)

    vec = np.asarray(np.gradient(gsdf))
    vec = vec.reshape(dimensions, -1)
    vec = batch_normalize(vec)

    e1 = vec.copy()
    e1[0] = -vec[1]
    e1[1] = vec[0]
    e1[2] = 0
    e1 = batch_normalize(e1)
    e2 = np.cross(vec.T, e1.T).T

    if sign is None or isinstance(sign, float):
        if isinstance(sign, float):
            s = compute_crossings_2d(gsdf[:, :, 0], thr=abs(sign))
        else:
            s = compute_crossings_2d(gsdf[:, :, 0])
        sign = np.repeat(s[:, :, np.newaxis], gsdf.shape[2], axis=2).flatten()

    qq = np.clip(pp, 0, 1)
    phis = sign * (+qq * np.pi + np.pi / 2)
    vec2 = rotate_vectors_axis(vec, e2, phis)

    o = np.zeros((3, uu.shape[0]))
    alpha = -np.arctan2(2 * ww / (d ** 2), 8 * uu / (w ** 2))
    o[:dimensions, :] = rotate_vectors_axis(vec2, e1, -2*alpha)
    o = batch_normalize(o)

    return o




# Copyright (C) 2024 Peter Ropač
# This file is part of SPOMSO.
# SPOMSO is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# SPOMSO is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public License along with SPOMSO. If not, see <https://www.gnu.org/licenses/>.

from .combine import CombineGeometry, smoothmax_boltz, smoothmin_poly2, smoothmin_poly3
from .transformations import EuclideanTransform, EuclideanTransformPoints
from .modifications import ModifyObject, ModifyVectorObject
from .geom import GenericGeometry, Points
from .helper_functions import resolution_conversion, generate_grid
from .helper_functions import smarter_reshape, vector_smarter_reshape, nd_vector_smarter_reshape
from .post_processing import sigmoid_falloff, positive_sigmoid_falloff, capped_exponential
from .post_processing import linear_falloff, relu, smooth_relu, slowstart
from .post_processing import hard_binarization
from .post_processing import gaussian_boundary, gaussian_falloff
from .post_processing import conv_averaging, conv_edge_detection
from .post_processing import custom_post_process
from .post_processing import PostProcess

from .sdf_2D import sdf_circle, sdf_segment_2d, sdf_box_2d, sdf_rounded_box_2d, sdf_triangle_2d
from .sdf_2D import sdf_arc, sdf_sector, sdf_inf_sector, sdf_ngon
from .sdf_2D import sdf_segmented_curve_2d, sdf_segmented_line_2d, sdf_parametric_curve_2d
from .sdf_2D import sdf_point_cloud_2d

from .geom_2d import Circle, NGon, Rectangle, RoundedRectangle, Segment, Triangle, Sector, InfiniteSector, Arc
from .geom_2d import ParametricCurve, SegmentedParametricCurve, SegmentedLine
from .geom_2d import PointCloud2D

from .sdf_3D import sdf_sphere, sdf_cylinder, sdf_box, sdf_torus, sdf_chainlink, sdf_braid, sdf_arc_3d
from .sdf_3D import sdf_plane, sudf_plane, sdf_segment_3d
from .sdf_3D import sdf_cone, sdf_oriented_infinite_cone, sdf_infinite_cone, sdf_solid_angle
from .sdf_3D import sdf_triangle_3d, sdf_quad_3d
from .sdf_3D import sdf_segmented_line_3d, sdf_segmented_curve_3d, sdf_parametric_curve_3d
from .sdf_3D import sdf_x, sdf_y, sdf_z
from .sdf_3D import sdf_point_cloud_3d

from .geom_3d import InfiniteCylinder, Cylinder, Sphere, Box, Plane, OrientedPlane, Line, Triangle3D, Quad
from .geom_3d import Torus, ChainLink, Braid, Arc3D, Cone, InfiniteCone, OrientedInfiniteCone
from .geom_3d import ParametricCurve3D, SegmentedParametricCurve3D, SegmentedLine3D
from .geom_3d import X, Y, Z

from .vector_functions import cartesian_define, spherical_define, cylindrical_define
from .vector_functions import radial_vector_field_spherical
from .vector_functions import radial_vector_field_cylindrical, hyperbolic_vector_field_cylindrical
from .vector_functions import aar_vector_field_cylindrical, awn_vector_field_cylindrical
from .vector_functions import vortex_vector_field_cylindrical, aav_vector_field_cylindrical
from .vector_functions import x_vector_field, y_vector_field, z_vector_field
from .vector_functions import from_sdf
from .geom_vector_special import lcwg1_p1, lcwg1_m1, lcwg1_2d
from .vector_functions_special import compute_crossings_2d

from .vector_modification_functions import batch_normalize
from .vector_modification_functions import add_vectors, subtract_vectors, rescale_vectors
from .vector_modification_functions import rotate_vectors_phi, rotate_vectors_theta
from .vector_modification_functions import rotate_vectors_x_axis, rotate_vectors_y_axis, rotate_vectors_z_axis
from .vector_modification_functions import rotate_vectors_axis
from .vector_modification_functions import revolve_field_x, revolve_field_y, revolve_field_z

from .geom_vector import CartesianVectorField, CylindricalVectorField, SphericalVectorField
from .geom_vector import RadialSphericalVectorField
from .geom_vector import RadialCylindricalVectorField, HyperbolicCylindricalVectorField
from .geom_vector import AngledRadialCylindricalVectorField, WindingCylindricalVectorField
from .geom_vector import VortexCylindricalVectorField, AngledVortexCylindricalVectorField
from .geom_vector import XVectorField, YVectorField, ZVectorField
from .geom_vector import VectorFieldFromSDF
from .geom_vector_special import LCWG2D, LCWG3Dp1, LCWG3Dm1



# Copyright (C) 2024 Peter Ropač
# This file is part of SPOMSO.
# SPOMSO is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# SPOMSO is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public License along with SPOMSO. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from scipy.interpolate import NearestNDInterpolator
from scipy.spatial.distance import cdist
from scipy.spatial import KDTree


def sdf_x(co, offset):
    return co[0] - offset


def sdf_y(co, offset):
    return co[1] - offset


def sdf_z(co, offset):
    return co[2] - offset


def sdf_sphere(co, radius):
    length = np.linalg.norm(co, axis=0)
    return length - radius


def sdf_cylinder(co, radius, height):
    l1 = np.linalg.norm(co[:2,:], axis=0)

    d0 = l1 - radius
    d1 = np.abs(co[2,:]) - height/2
    term1 = np.minimum(np.maximum(d0, d1), 0)
    term2 = np.linalg.norm([np.maximum(d0, 0), np.maximum(d1, 0)], axis=0)
    return term1 + term2


def sdf_box(co, a, b, c):

    v = np.asarray((a,b,c))
    qo = np.abs(co).T - v

    t1 = np.linalg.norm(np.maximum(qo, 0.0), axis=1)
    t2 = np.minimum(np.maximum(qo[:, 0], np.maximum(qo[:, 1], qo[:, 2])), 0.0)
    return t1 + t2


def sdf_torus(co, R, r):
    p = np.linalg.norm(co[:2], axis=0) - R
    q = np.linalg.norm([p, co[2]], axis=0)
    return q - r


def sdf_chainlink(co, R, r, length):
    co[0] = co[0] - np.clip(co[0], -length/2, length/2)
    p = np.linalg.norm(co[:2], axis=0) - R
    q = np.linalg.norm([p, co[2]], axis=0)
    return q - r


def sdf_braid(co, length, R, r, pitch):
    c = np.cos(pitch * co[2])
    s = np.sin(pitch * co[2])
    rot = np.asarray([[c, s], [-s, c]])
    co[:2, :] = rot[0, :, :] * co[0, :] + rot[1, :, :] * co[1, :]

    co[2] -= np.clip(co[2], -length / 2, length / 2)

    p = np.linalg.norm(co[[0,2]], axis=0) - R
    q = np.linalg.norm([p, co[1]], axis=0)
    return q - r


def sdf_arc_3d(co, R, r, start_angle, end_angle):

    c_angle = (start_angle + end_angle) / 2
    rot = np.asarray([[np.cos(c_angle), np.sin(c_angle)], [-np.sin(c_angle), np.cos(c_angle)]])
    co[:2] = rot.dot(co[:2])
    co[1] = np.abs(co[1])
    phi = np.arctan2(co[1], co[0])

    end_angle = np.abs(end_angle - c_angle)

    psi = np.clip(phi, 0, end_angle)
    oc = np.asarray((R * np.cos(psi),
                     R * np.sin(psi)))

    co[:2] = np.subtract(co[:2], oc)

    length = np.linalg.norm(co, axis=0) - r
    return length


def sdf_plane(co, normal, offset):
    n = normal/np.linalg.norm(normal)
    d = np.dot(co.T, n).T
    return d - offset


def sudf_plane(co, normal, thickness):
    n = normal/np.linalg.norm(normal)
    d = np.abs(np.dot(co.T, n).T)
    return d - thickness


def sdf_segment_3d(co, a, b):
    a = np.asarray(a)
    b = np.asarray(b)
    pa = np.subtract(co.T, a).T
    ba = b - a
    h = np.clip(np.sum(np.multiply(pa.T, ba).T, axis=0)/np.dot(ba, ba), 0, 1)
    length = np.linalg.norm(pa - np.outer(ba,h), axis=0)
    return length


def sdf_cone(co, height, angle):
    # https://iquilezles.org/articles/distfunctions/
    q = np.asarray((np.tan(angle), -1))*height
    co[2] -= height*(0.5**(1/3))

    w = np.asarray((np.linalg.norm(co[:2], axis=0), co[2]))
    ones = np.ones(w.shape[1])
    a = w - np.outer(q, np.clip(np.dot(w.T, q).T / np.dot(q, q), 0.0, 1.0))
    b = w.T - q*np.asarray((np.clip(w[0] / q[0], 0.0, 1.0), ones)).T

    aa = np.sum(np.multiply(a,a), axis=0)
    bb = np.sum(np.multiply(b,b), axis=1)
    d = np.minimum(aa, bb)
    s = np.maximum(-(w[0] * q[1] - w[1] * q[0]), -(w[1] - q[1]))
    return np.sqrt(d) * np.sign(s)


def sdf_oriented_infinite_cone(co, angle):
    # https://iquilezles.org/articles/distfunctions/
    v = np.asarray([np.sin(angle), np.cos(angle)])
    q = np.asarray((np.linalg.norm(co[:2], axis=0), -co[2]))
    qv = np.outer(v, np.maximum(np.dot(q.T, v).T, 0.0))
    d = np.linalg.norm(q - qv)

    f = -2*(q[0] * v[1] - q[1] * v[0] < 0.0) + 1

    return d * f


def sdf_infinite_cone(co, angle):
    v = np.asarray([np.sin(angle), np.cos(angle)])
    q = np.asarray((np.linalg.norm(co[:2], axis=0), -co[2]))
    qv = np.outer(v, np.maximum(np.dot(q.T, v).T, 0.0))
    d = np.linalg.norm(q - qv)

    f = -2 * (q[0] * v[1] - q[1] * v[0] < 0.0) + 1

    return np.abs(d * f)


def sdf_solid_angle(co, radius, angle_1, angle_2):
    angle_diff = np.abs(angle_2 - angle_1)/2
    c_angle = (angle_2 + angle_1)/2
    rot = np.asarray([[np.cos(c_angle), np.sin(c_angle)], [-np.sin(c_angle), np.cos(c_angle)]])
    co[:2] = rot.dot(co[:2])
    co[1] = np.linalg.norm(co[1:,:], axis=0)

    phi = np.arctan2(co[1], co[0])
    psi = np.clip(phi, 0, angle_diff)
    oc = np.asarray((radius * np.cos(psi),
                     radius * np.sin(psi)))
    poc = np.subtract(co[:2], oc)
    length = np.linalg.norm(poc, axis=0)

    c = np.asarray([np.cos(angle_diff), np.sin(angle_diff)])
    outer = np.outer(c, np.clip( np.dot(co[:2].T, c).T, 0, radius))
    m = np.linalg.norm(co[:2] - outer, axis=0)

    msk = (np.linalg.norm(co[:2], axis=0) <= radius)*(phi <= angle_diff)
    out = np.minimum(m,length)
    out[msk] = -out[msk]

    return out


def sdf_triangle_3d(co, a, b, c):
    # https://iquilezles.org/articles/distfunctions/
    s1 = (b-a)/1
    s2 = (c-b)/1
    s3 = (a-c)/1
    
    coa = np.subtract(co.T, a).T
    cob = np.subtract(co.T, b).T
    coc = np.subtract(co.T, c).T

    normal = np.cross(s1, s3)

    mask = np.sign(np.dot(np.cross(s1, normal), coa)) + np.sign(np.dot(np.cross(s2, normal), cob))
    mask += np.sign(np.dot(np.cross(s3, normal), coc))
    mask = mask < 2.

    term1 = np.outer(s1, np.clip(np.dot(s1, coa[:, mask]) / np.dot(s1, s1), 0, 1) ) - coa[:, mask]
    term2 = np.outer(s2, np.clip(np.dot(s2, cob[:, mask]) / np.dot(s2, s2), 0, 1)) - cob[:, mask]
    term3 = np.outer(s3, np.clip(np.dot(s3, coc[:, mask]) / np.dot(s3, s3), 0, 1)) - coc[:, mask]
    ex1 = np.minimum( np.minimum(np.sum(term1*term1, axis=0), np.sum(term2*term2, axis=0)), np.sum(term3*term3, axis=0) )

    mask2 = np.invert(mask)
    ex2 = (np.dot(normal, coa[:, mask2])**2)/np.dot(normal, normal)

    outer = np.zeros(co.shape[1])
    outer[mask] = np.sqrt(ex1)
    outer[mask2] = np.sqrt(ex2)

    return outer


def sdf_quad_3d(co, a, b, c, d):
    # https://iquilezles.org/articles/distfunctions/
    s1 = (b - a)
    s2 = (c - b)
    s3 = (d - c)
    s4 = (a - d)

    coa = np.subtract(co.T, a).T
    cob = np.subtract(co.T, b).T
    coc = np.subtract(co.T, c).T
    cod = np.subtract(co.T, d).T

    normal = np.cross(s1, s4)

    mask = np.sign(np.dot(np.cross(s1, normal), coa)) + np.sign(np.dot(np.cross(s2, normal), cob))
    mask += np.sign(np.dot(np.cross(s3, normal), coc)) + np.sign(np.dot(np.cross(s4, normal), cod))
    mask = mask < 3.

    term1 = np.outer(s1, np.clip(np.dot(s1, coa[:, mask]) / np.dot(s1, s1), 0, 1)) - coa[:, mask]
    term2 = np.outer(s2, np.clip(np.dot(s2, cob[:, mask]) / np.dot(s2, s2), 0, 1)) - cob[:, mask]
    term3 = np.outer(s3, np.clip(np.dot(s3, coc[:, mask]) / np.dot(s3, s3), 0, 1)) - coc[:, mask]
    term4 = np.outer(s4, np.clip(np.dot(s4, cod[:, mask]) / np.dot(s4, s4), 0, 1)) - cod[:, mask]
    s_ex1 = np.minimum(np.sum(term1 * term1, axis=0), np.sum(term2 * term2, axis=0))
    s_ex2 = np.minimum(np.sum(term4 * term4, axis=0), np.sum(term3 * term3, axis=0))
    ex1 = np.minimum(s_ex2, s_ex1)

    mask2 = np.invert(mask)
    ex2 = (np.dot(normal, coa[:, mask2]) ** 2) / np.dot(normal, normal)

    outer = np.zeros(co.shape[1])
    outer[mask] = np.sqrt(ex1)
    outer[mask2] = np.sqrt(ex2)

    return outer


def sdf_segmented_curve_3d(co, points, t):

    v = np.floor(t).astype(int)
    u = t - v
    fval = points[:, v+1]*u + points[:, v]*(1-u)

    tree = KDTree(fval.T)
    mindist, minid = tree.query(co[:, :].T)
    return mindist


def sdf_segmented_line_3d(co, points):

    out = np.ones(co.shape[1])*10000
    for i in range(points.shape[1]-1):
        part = sdf_segment_3d(co, points[:, i], points[:, i+1])
        out = np.minimum(out, part)

    return out


def sdf_parametric_curve_3d(co, f, f_parameters, t):

    fval = f(t, *f_parameters)

    tree = KDTree(fval.T)
    mindist, minid = tree.query(co[:, :].T)
    return mindist


def sdf_point_cloud_3d(co, points):
    tree = KDTree(points.T)
    mindist, minid = tree.query(co[:3, :].T)
    return mindist




# Copyright (C) 2024 Peter Ropač
# This file is part of SPOMSO.
# SPOMSO is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# SPOMSO is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public License along with SPOMSO. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from scipy.spatial.transform import Rotation as srot
from typing import Callable


class EuclideanTransform:
    """
    Class containing the Euclidian transforms which can be applied to a geometry.
    """

    def __init__(self):
        self._et: list = []
        self._center: np.ndarray | tuple | list = np.asarray((0.0,0.0,0.0))
        self._scale: float | int = 1.0
        self._rot_matrix: np.ndarray = np.eye(3)
        self._angle: float | int = 0.0
        self._axis: np.ndarray | tuple | list = np.asarray((0.0, 0.0, 1.0))

    @property
    def transformations(self) -> list:
        """All the Euclidian transformations which were applied to the geometry in chronological order.

        Returns:
            List of Euclidian transformations.
        """
        return self._et

    @property
    def center(self) -> np.ndarray | tuple | list:
        """Center of mass of the geometry.

        Returns:
            Position vector of the center of mass of the geometry.
        """
        return self._center

    @property
    def scale(self) -> float | list:
        """Scale factor of the geometry.

        Returns:
            Scale factor.
        """
        return self._scale

    @property
    def rotation_matrix(self) -> np.ndarray:
        """Rotation matrix applied to the geometry.

        Returns:
            Rotation matrix (3, 3).
        """
        return self._rot_matrix

    @property
    def rotation_axis(self) -> np.ndarray | tuple | list:
        """Vector representing the axis of rotation.

        Returns:
            Axis of rotation.
        """
        return self._axis

    @property
    def rotation_angle(self) -> float | int:
        """Angle by which the geometry is rotated around the axis of rotation.

        Returns:
           Angle of rotation.
        """
        return self._angle

    def set_location(self, center: np.ndarray | tuple | list):
        """Set the position of the geometry in 3D space.
        
        Args:
            center: 3vector which defines the new position of the geometry.
        """
        self._et.append("set_location")

        center = np.asarray(center)
        if center.size <= 3:
            self._center[:center.size] = center
        else:
            raise SyntaxError(f"Array {center} is of incorrect size!")

    def move(self, move_vector: np.ndarray | tuple | list):
        """
        Move the geometry by a vector.
        
        Args:
            move_vector: 3vector by which the geometry is moved.
        """
        self._et.append("move")

        vector = np.asarray(move_vector)
        if vector.size <= 3:
            self._center += vector
        else:
            raise SyntaxError(f"Array {vector} is of incorrect size!")

    def set_scale(self, scale: float | int):
        """
        Sets the scaling of the geometry.
        
        Args:
            scale: Scaling factor.
        """
        self._et.append("set_scale")

        if any(isinstance(scale, type_) for type_ in [float, int]):
            self._scale = scale
        else:
            raise TypeError("Scale must be a float or an int")

    def rescale(self, scale: float | int):
        """
        Multiplies the existing scaling of the geometry by the scale factor.
        
        Args:
            scale: Scale factor.
        """
        self._et.append("rescale")

        if any(isinstance(scale, type_) for type_ in [float, int]):
            self._scale *= scale
        else:
            raise TypeError("Scale must be a float or an int")

    @staticmethod
    def get_rotation_matrix(angle: float | int, axis: np.ndarray | tuple | list) -> tuple:
        """
        Get the rotation matrix from the rotation angle and axis of rotation.
        
        Args:
            angle: Angle of rotation.
            axis: Axis of rotation.
        
        Returns:
            Rotation matrix (3, 3).
        """
        axis = np.asarray(axis)
        if axis.size <= 3:
            axis_ = np.zeros(3)
            axis_[:axis.size] = axis
        else:
            raise SyntaxError(f"Array {axis} is of incorrect size!")

        if any(isinstance(angle, type_) for type_ in [float, int]):
            angle_ = angle
        else:
            raise TypeError("Rotation angle must be a float or an int")

        r = srot.from_rotvec(angle_ * axis_)
        return r.as_matrix(), angle_, axis_

    def set_rotation(self, angle: float | int, axis: np.ndarray | tuple | list):
        """
        Sets the rotation matrix from the rotation angle and axis.
        
        Args:
            angle: Angle of rotation.
            axis: Axis of rotation.
        """
        self._et.append("set_rotation")

        self._rot_matrix, self._angle, self._axis = self.get_rotation_matrix(angle, axis)

    def rotate_rotvec(self, angle: float | int, axis: np.ndarray | tuple | list):
        """
        Multiplies the previous existing rotation matrix by a rotation matrix calculated
        from the angle and axis of rotation.
        
        Args:
            angle: Angle of rotation.
            axis: Axis of rotation.
        """
        de = np.array_equal(axis, np.zeros(3)[:axis.size])
        if de:
            raise ValueError("Axis cannot be zero!")

        axis = axis/np.linalg.norm(axis)
        rot_matrix, angle_, axis_ = self.get_rotation_matrix(angle, axis)

        self.rotate_matrix(rot_matrix)

    def rotate_matrix(self, rotation_matrix: np.ndarray | tuple | list):
        """
        Multiplies the previous existing rotation matrix by a given rotation matrix.
        
        Args:
            rotation_matrix: Rotation matrix (3, 3).
        """
        self._rot_matrix = np.matmul(rotation_matrix, self._rot_matrix)

        r = srot.from_matrix(self._rot_matrix)
        rot_vec_ = r.as_rotvec()
        self._angle = np.linalg.norm(rot_vec_)
        if not self._angle == 0:
            self._axis = rot_vec_ / self._angle
        else:
            self._axis = np.asarray((0.0, 0.0, 1.0))

    def rotate(self, *inputs: np.ndarray | tuple):
        """
        Rotates the geometry by a rotation matrix or a rotation vector.
        
        Args:
            inputs: Rotation metrix or angle and axis vector.
        """
        self._et.append("rotate")

        n_inputs = len(inputs)

        if n_inputs==1:
            rot_matrix = np.asarray(inputs)
            self.rotate_matrix(rot_matrix)

        if n_inputs==2:
            angle, axis = inputs
            self.rotate_rotvec(angle, np.asarray(axis))

        if n_inputs>2:
            raise SyntaxError("Wrong number of inputs!")

    @staticmethod
    def apply_ec_transforms(function_: Callable[[np.ndarray, tuple], np.ndarray],
                            co_: np.ndarray,
                            params_: tuple,
                            rm: np.ndarray, tm: np.ndarray, sm: float | int) -> np.ndarray:
        rm = rm.T
        co = rm.dot(co_)
        co = co/sm
        co = np.subtract(co.T, rm.dot(tm)).T

        return sm*function_(co, *params_)

    def apply(self,
              function_: Callable[[np.ndarray, tuple], np.ndarray],
              co_: np.ndarray,
              params_: tuple) -> np.ndarray:
        """
        Apply the transformations to the geometry (SDF).
        
        Args:
            function_: Original SDF
            co_:  Point cloud of coordinates with shape (D, N);
                D - number of dimensions (2 or 3);
                N - number of points in the point cloud.
            params_: Parameters of the SDF.
        
        Returns:
            Signed Distance field of shape (N,).
        """
        return self.apply_ec_transforms(function_, co_, params_,
                                        self.rotation_matrix,
                                        self.center,
                                        self.scale)


class EuclideanTransformPoints:
    """
    Class containing the Euclidian transforms which can be applied to a point cloud.
    """

    def __init__(self):
        self._et: list = []
        self._center: np.ndarray | tuple | list = np.asarray((0.0, 0.0, 0.0))
        self._scale: np.ndarray | tuple | list | float | int = 1.0
        self._rot_matrix: np.ndarray = np.eye(3)
        self._angle: float | int = 0.0
        self._axis: np.ndarray | tuple | list = np.asarray((0.0, 0.0, 1.0))

    @property
    def transformations(self) -> list:
        """All the Euclidian transformations which were applied to the point cloud in chronological order.

        Returns:
            List of Euclidian transformations.
        """
        return self._et

    @property
    def center(self) -> np.ndarray | tuple | list:
        """Center of mass of the point cloud.

        Returns:
            Position vector of the center of mass of the point cloud.
        """
        return self._center

    @property
    def scale(self) -> np.ndarray | tuple | list | float | int:
        """Scale factors of the point cloud.

        Returns:
            Scale factor.
        """
        return self._scale

    @property
    def rotation_matrix(self) -> np.ndarray:
        """Rotation matrix applied to the point cloud.

        Returns:
            Rotation matrix (3, 3).
        """
        return self._rot_matrix

    @property
    def rotation_axis(self) -> np.ndarray | tuple | list:
        """Vector representing the axis of rotation.

        Returns:
            Axis of rotation.
        """
        return self._axis

    @property
    def rotation_angle(self) -> float | int:
        """Angle by which the point cloud is rotated around the axis of rotation.

        Returns:
           Angle of rotation.
        """
        return self._angle

    def set_location(self, center: np.ndarray | tuple | list):
        """Set the position of the point cloud in 3D space.
        
        Args:
            center: 3vector which defines the new position of the point cloud.
        """
        self._et.append("set_location")

        center = np.asarray(center)
        if center.size <= 3:
            self._center[:center.size] = center
        else:
            raise SyntaxError(f"Array {center} is of incorrect size!")

    def move(self, move_vector: np.ndarray | tuple | list):
        """
        Move the point cloud by a vector.
        
        Args:
            move_vector: 3vector by which the point cloud is moved.
        """
        self._et.append("move")

        vector = np.asarray(move_vector)
        if vector.size <= 3:
            self._center += vector
        else:
            raise SyntaxError(f"Array {vector} is of incorrect size!")

    def set_scale(self, scale: np.ndarray | tuple | list | float | int):
        """
        Sets the scaling of the point cloud.
        
        Args:
            scale: Scaling factors.
        """
        self._et.append("set_scale")

        if any(isinstance(scale, type_) for type_ in [float, int]):
            self._scale = scale*np.ones(3)
        elif any(isinstance(scale, type_) for type_ in [np.ndarray, tuple, list]):
            scale_ = np.ones(3)
            scale = np.asarray(scale)
            scale_[:scale.size] = scale
            self._scale = scale_
        else:
            raise TypeError("Wrong data type, try a 3vector (np.ndarray, tuple, list) or a scalar.")

    def rescale(self, scale: np.ndarray | tuple | list | float | int):
        """
        Multiplies the existing scaling of the point cloud by the scale factors.
        
        Args:
            scale: Scale factors.
        """
        self._et.append("rescale")

        if any(isinstance(scale, type_) for type_ in [float, int]):
            self._scale *= scale*np.ones(3)
        elif any(isinstance(scale, type_) for type_ in [np.ndarray, tuple, list]):
            self._scale = np.multiply(self._scale, scale)
        else:
            raise TypeError("Wrong data type, try a 3vector (np.ndarray, tuple, list) or a scalar.")

    @staticmethod
    def get_rotation_matrix(angle: float | int, axis: np.ndarray | tuple | list) -> tuple:
        """
        Get the rotation matrix from the rotation angle and axis of rotation.
        
        Args:
            angle: Angle of rotation.
            axis: Axis of rotation.
        
        Returns:
            Rotation matrix (3, 3).
        """
        axis = np.asarray(axis)
        if axis.size <= 3:
            axis_ = np.zeros(3)
            axis_[:axis.size] = axis
        else:
            raise SyntaxError(f"Array {axis} is of incorrect size!")

        if any(isinstance(angle, type_) for type_ in [float, int]):
            angle_ = angle
        else:
            raise TypeError("Rotation angle must be a float or an int.")

        r = srot.from_rotvec(angle_ * axis_)
        return r.as_matrix(), angle_, axis_

    def set_rotation(self, angle: float | int, axis: np.ndarray | tuple | list):
        """
        Sets the rotation matrix from the rotation angle and axis.
        
        Args:
            angle: Angle of rotation.
            axis: Axis of rotation.
        """
        self._et.append("set_rotation")

        self._rot_matrix, self._angle, self._axis = self.get_rotation_matrix(angle, axis)

    def rotate_rotvec(self, angle: float | int, axis: np.ndarray | tuple | list):
        """
        Multiplies the previous existing rotation matrix by a rotation matrix calculated
        from the rotation angle and axis of rotation.
        
        Args:
            angle: Angle of rotation.
            axis: Axis of rotation.
        """
        de = np.array_equal(axis, np.zeros(3)[:axis.size])
        if de:
            raise ValueError("Axis cannot be zero!")

        axis = axis/np.linalg.norm(axis)
        rot_matrix, angle_, axis_ = self.get_rotation_matrix(angle, axis)

        self.rotate_matrix(rot_matrix)

    def rotate_matrix(self, rotation_matrix: np.ndarray | tuple | list):
        """
        Multiplies the previous existing rotation matrix by a given rotation matrix.
        
        Args:
            rotation_matrix: Rotation matrix (3, 3).
        """
        self._rot_matrix = np.matmul(rotation_matrix, self._rot_matrix)

        r = srot.from_matrix(self._rot_matrix)
        rot_vec_ = r.as_rotvec()
        self._angle = np.linalg.norm(rot_vec_)
        if not self._angle == 0:
            self._axis = rot_vec_ / self._angle
        else:
            self._axis = np.asarray((0.0, 0.0, 1.0))

    def rotate(self, *inputs: np.ndarray | tuple):
        """
        Rotates by a rotation matrix or a rotation vector.
        
        Args:
            inputs: Rotation metrix or angle and axis vector.
        """
        self._et.append("rotate")

        n_inputs = len(inputs)

        if n_inputs==1:
            rot_matrix = np.asarray(inputs)
            self.rotate_matrix(rot_matrix)

        if n_inputs==2:
            angle, axis = inputs
            self.rotate_rotvec(angle, np.asarray(axis))

        if n_inputs>2:
            raise SyntaxError("Wrong number of inputs!")

    @staticmethod
    def apply_ec_transforms(points_: np.ndarray,
                            rm: np.ndarray, tm: np.ndarray, sm: np.ndarray) -> np.ndarray:

        co = np.multiply(points_.T, sm).T
        rm = rm.T
        co = rm.dot(co)
        co = np.subtract(co.T, rm.dot(tm)).T

        return co

    def apply(self, points_: np.ndarray) -> np.ndarray:
        """
        Apply the transformations to the point cloud.
        
        Args:
            points_: Point cloud of shape (D, N)
        
        Returns:
            Transformed point cloud of shape (D, N).
        """
        return self.apply_ec_transforms(points_,
                                        self.rotation_matrix,
                                        self.center,
                                        self.scale)